from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Optional, TYPE_CHECKING
from enum import Enum

from mdk.types.specifier import TypeSpecifier

## an ugly hack, necessary due to circular import.
if TYPE_CHECKING:
    from mdk.types.expressions import Expression
    from mdk.resources.animation import Animation

class TranslationMode(Enum):
    STANDARD = 0
    VARIABLE = 1

class StateScopeType(Enum):
    SHARED = 0
    PLAYER = 1
    HELPER = 2
    TARGET = 3

@dataclass
class StateScope:
    scope: StateScopeType
    target: Optional[int] = None
    def __str__(self) -> str:
        if self.scope == StateScopeType.SHARED:
            return "shared"
        elif self.scope == StateScopeType.PLAYER:
            return "player"
        elif self.scope == StateScopeType.TARGET:
            return "target"
        elif self.scope == StateScopeType.HELPER:
            if self.target == None: return "helper"
            return f"helper({self.target})"
        return "<?>"
    
SCOPE_TARGET = StateScope(StateScopeType.TARGET)
SCOPE_PLAYER = StateScope(StateScopeType.PLAYER)
SCOPE_HELPER: Callable[[Optional[int]], StateScope] = lambda id: StateScope(StateScopeType.HELPER, id)

@dataclass
class StateController:
    type: str
    params: dict[str, Expression]
    triggers: list[Expression]
    location: tuple[str, int]

    def __init__(self):
        self.type = ""
        self.params = {}
        self.triggers = []
        self.location = ("<?>", 0)

    def __repr__(self):
        result = "[State ]"
        result += f"\ntype = {self.type}"
        for trigger in self.triggers:
            result += f"\ntrigger1 = {trigger.__str__()}"
        for param in self.params:
            result += f"\n{param} = {self.params[param].__str__()}"
        return result
    
@dataclass
class ParameterDefinition:
    type: TypeSpecifier
    name: str
    ## this is only supported for globals.
    ## by default globals use SHARED scope, but we should be able to specify.
    scope: Optional[StateScope] = None
    is_system: bool = False

@dataclass
class StateDefinition:
    fn: Callable[[], None]
    params: dict[str, Expression]
    controllers: list[StateController]
    locals: list[ParameterDefinition]
    location: tuple[str, int]
    scope: Optional[StateScope]
    _fwd_animation: Optional[Animation]

@dataclass
class TemplateDefinition:
    fn: Callable[..., None]
    library: Optional[str]
    params: dict[str, TypeSpecifier]
    controllers: list[StateController]
    locals: list[ParameterDefinition]

@dataclass
class TriggerDefinition:
    fn: Callable[..., Expression]
    library: Optional[str]
    result: TypeSpecifier
    params: dict[str, TypeSpecifier]
    exprn: Optional[Expression] = None

@dataclass
class CompilerContext:
    statedefs: dict[str, StateDefinition]
    templates: dict[str, TemplateDefinition]
    triggers: dict[str, TriggerDefinition]
    typedefs: dict[str, TypeSpecifier]
    current_state: Optional[StateDefinition]
    current_template: Optional[TemplateDefinition]
    current_trigger: Optional[Expression]
    trigger_stack: list[Expression]
    check_stack: list[Expression]
    globals: list[ParameterDefinition]
    default_state: tuple[Expression | None, Expression | None]
    statefuncs: list[str]
    format_params: list[Expression]
    if_stack: list[int]
    animations: list[Animation]
    debug_build: bool

    def __init__(self):
        self.statedefs = {}
        self.templates = {}
        self.triggers = {}
        self.typedefs = {}
        self.current_state = None
        self.current_template = None
        self.current_trigger = None
        self.trigger_stack = []
        self.check_stack = []
        self.globals = []
        self.default_state = (None, None)
        self.statefuncs = []
        self.format_params = []
        self.if_stack = []
        self.animations = []
        self.debug_build = False
    
    @classmethod
    def instance(cls):
        if not hasattr(cls, '_instance'):
            cls._instance = CompilerContext()
        return cls._instance
    
    def get_next_anim_id(self) -> int:
        """Gets the next free animation number, starting from 0."""
        index = 0
        ls = sorted([a for a in self.animations if a._id != None], key = lambda x: x._id or 0)
        for anim in ls:
            if anim._id == index:
                index += 1
            else:
                return index
        return index



## contains stdlib functions which are exposed by MTL (non-CNS stdlib)
__all__ = ["rescope"]

from mdk.types import StateScope

from mdk.stdlib.redirects import RedirectTarget

def rescope(source: RedirectTarget, target: StateScope) -> RedirectTarget:
    new_target = str(target)
    if new_target == "player": new_target = "root"
    if new_target == "shared": raise Exception(f"target parameter to rescope should not be SHARED.")
    return RedirectTarget(f"rescope({source.__repr__()}, {new_target})")


import functools
from typing import Union, Callable
from enum import Enum, Flag

from mdk.types.specifier import TypeSpecifier, TypeCategory
from mdk.types.builtins import IntType, BoolType, FloatType, StringType, StateNoType, AnyType
from mdk.types.context import CompilerContext, StateController

from mdk.utils.expressions import check_types_assignable

## this is just the 'convert' function from mdk.utils.shared,
## but copied here to avoid circular imports.
def _convert(input: Union['Expression', Enum, Flag, str, int, float, bool, tuple, Callable[..., StateController]]) -> 'Expression':
    if isinstance(input, Expression):
        return input
    elif isinstance(input, functools.partial):
        return Expression(input.keywords["value"], StateNoType)
    elif isinstance(input, Callable):
        return Expression(input.__name__, StateNoType)
    elif isinstance(input, tuple):
        elements: list[str] = []
        for elem in input:
            elements.append(_convert(elem).exprn)
        return Expression(", ".join(elements), AnyType)
    ## cursed check to avoid circular import on Animation
    elif type(input).__name__ == "Animation":
        if input._id == None: # type: ignore
            raise Exception("Animation without an assigned ID cannot be used as an Expression! (Did you try to use an automatic-numbered Animation in global state?)")
        return Expression(str(input._id), IntType) # type: ignore
    elif type(input) == str:
        return Expression(input, StringType)
    elif type(input) == int:
        return Expression(str(input), IntType)
    elif type(input) == float:
        return Expression(str(input), FloatType)
    elif type(input) == bool:
        return Expression("true" if input else "false", BoolType)
    elif isinstance(input, Flag):
        ctx = CompilerContext.instance()
        for typename in ctx.typedefs:
            typedef = ctx.typedefs[typename]
            if hasattr(typedef, "inner_type") and typedef.inner_type == type(input): # type: ignore
                if not typedef.register:
                    result = ""
                    for member in input:
                        result += member.name # type: ignore
                    return Expression(f"{typedef.name}.{result}", typedef)
                else:
                    result = " | ".join([x.__str__() for x in input])
                    return Expression(f"({result})", typedef)
        raise Exception(f"Could not determine the MTL type to use for flag type {type(input)}.")
    elif isinstance(input, Enum):
        ctx = CompilerContext.instance()
        for typename in ctx.typedefs:
            typedef = ctx.typedefs[typename]
            if hasattr(typedef, "inner_type") and typedef.inner_type == type(input): # type: ignore
                return Expression(f"{typedef.name}.{input.name}", typedef)
        raise Exception(f"Could not determine the MTL type to use for enum type {type(input)}.")
    else:
        raise Exception(f"Attempted to convert from unsupported builtin type {type(input)}.")

## an Expression is a core component of building CNS.
## Expressions represent all trigger expressions, parameter expressions, etc.
class Expression:
    exprn: str
    type: TypeSpecifier
    def __init__(self, exprn: str, type: TypeSpecifier):
        self.exprn = exprn
        self.type = type
    def __repr__(self):
        return self.exprn
    def __str__(self):
        if self.type == StringType:
            return f'"{self.exprn}"'
        return self.exprn
    
    def make_expression(self, exprn: str):
        return Expression(exprn, self.type)
    
    ## comparisons
    def __eq__(self, other): # type: ignore
        if other is None:
            return NotImplemented
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot be compared.")
        return Expression(f"({self.__str__()} = {other.__str__()})", BoolType)
    def __ne__(self, other): # type: ignore
        if other is None:
            return NotImplemented
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot be compared.")
        return Expression(f"({self.__str__()} != {other.__str__()})", BoolType)
    def __lt__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot be compared.")
        return Expression(f"({self.__str__()} < {other.__str__()})", BoolType)
    def __le__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot be compared.")
        return Expression(f"({self.__str__()} <= {other.__str__()})", BoolType)
    def __gt__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot be compared.")
        return Expression(f"({self.__str__()} > {other.__str__()})", BoolType)
    def __ge__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot be compared.")
        return Expression(f"({self.__str__()} >= {other.__str__()})", BoolType)
    
    ## mathematical operations
    def __add__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot be added.")
        return Expression(f"({self.__str__()} + {other.__str__()})", self.type)
    def __sub__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot be subtracted.")
        return Expression(f"({self.__str__()} - {other.__str__()})", self.type)
    def __mul__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot be multiplied.")
        return Expression(f"({self.__str__()} * {other.__str__()})", self.type)
    def __truediv__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot be divided.")
        return Expression(f"({self.__str__()} / {other.__str__()})", self.type)
    def __floordiv__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot be divided.")
        return Expression(f"floor({self.__str__()} / {other.__str__()})", self.type)
    def __mod__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot have modulus taken.")
        ## TODO: confirm the input types are both `int` or assignable to `int`
        return Expression(f"({self.__str__()} % {other.__str__()})", IntType)
    def __pow__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot have exponentiation taken.")
        return Expression(f"({self.__str__()} ** {other.__str__()})", self.type)
    def __rshift__(self, other):
        raise Exception()
    def __lshift__(self, other):
        raise Exception()
    def __and__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot have bitwise and taken.")
        return Expression(f"({self.__str__()} & {other.__str__()})", self.type)
    def __or__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot have bitwise or taken.")
        return Expression(f"({self.__str__()} | {other.__str__()})", self.type)
    def __xor__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot have bitwise xor taken.")
        return Expression(f"({self.__str__()} ^ {other.__str__()})", self.type)
    def __radd__(self, other): return self.__add__(other)
    def __rsub__(self, other): return self.__sub__(other)
    def __rmul__(self, other): return self.__mul__(other)
    def __rtruediv__(self, other): return self.__truediv__(other)
    def __rfloordiv__(self, other): return self.__floordiv__(other)
    def __rmod__(self, other): return self.__mod__(other)
    def __rpow__(self, other): return self.__pow__(other)
    def __rlshift__(self, other): return self.__lshift__(other)
    def __rrshift__(self, other): return self.__rshift__(other)
    def __rand__(self, other): return self.__and__(other)
    def __ror__(self, other): return self.__or__(other)
    def __rxor__(self, other): return self.__xor__(other)
    def __neg__(self):
        return Expression(f"-({self.__str__()})", self.type)
    def __pos__(self):
        return Expression(f"+({self.__str__()})", self.type)
    def __abs__(self):
        return Expression(f"abs({self.__str__()})", self.type)
    def __invert__(self):
        return Expression(f"~({self.__str__()})", self.type)
    def __round__(self):
        return Expression(f"floor({self.__str__()})", self.type)
    def __trunc__(self):
        return Expression(f"floor({self.__str__()})", self.type)
    def __floor__(self):
        return Expression(f"floor({self.__str__()})", self.type)
    def __ceil__(self):
        return Expression(f"ceil({self.__str__()})", self.type)
    
    ## membership, for Flag types
    def __contains__(self, other):
        if not isinstance(other, Expression):
            other = _convert(other)
        if check_types_assignable(self.type, other.type) == None:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not assignable and cannot be tested for membership.")
        if self.type.category != TypeCategory.FLAG or other.type.category != TypeCategory.FLAG:
            raise Exception(f"Types {self.type.name} and {other.type.name} are not Flag types and cannot be tested for membership.")
        return Expression(f"{self.__str__()} = {other.__str__()}", BoolType)
    
    def __bool__(self):
        ctx = CompilerContext.instance()
        ctx.current_trigger = Expression(self.exprn, self.type)
        return True
    
    def __format__(self, spec: str):
        ## MUGEN valid specifiers are: %d, %i, %f, %F, %e, %E, %g, %G
        ## the only thing we need to support explicitly from here is e/E and g/G
        ## none of the other Python specifiers seem to be supported by MUGEN.
        if spec not in ["e", "E", "g", "G", ""]:
            raise Exception(f"Expected type specifier to be e, E, g, or G, but received {spec}. Other type specifies are not supported by MUGEN.")
        
        ## add this expression to the context
        ctx = CompilerContext.instance()
        ctx.format_params.append(self)

        if spec == "e":
            return "%e"
        elif spec == "E":
            return "%E"
        elif spec == "g":
            return "%g"
        elif spec == "G":
            return "%G"
        
        if self.type == FloatType:
            return "%f"
        else:
            return "%d"
    
## these are helpers for specifying expressions for commonly-used built-in types.
IntExpression = functools.partial(Expression, type = IntType)

## type hint for a tuple with variable length
type ConvertibleExpression = Union[Expression, str, int, float, bool, Enum, Flag]
type TupleExpression = tuple[ConvertibleExpression, ...]

__all__ = ["Expression", "IntExpression", "ConvertibleExpression", "TupleExpression"]


from mdk.types.builtins import *
from mdk.types.defined import *
from mdk.types.expressions import *
from mdk.types.specifier import TypeSpecifier
from mdk.types.variables import *
from mdk.types.enums import *
from mdk.types.context import StateScopeType, StateScope, SCOPE_HELPER, SCOPE_PLAYER, SCOPE_TARGET


from mdk.types.builtins import *
from dataclasses import dataclass
from mdk.types.expressions import Expression
from mdk.types.specifier import TypeCategory, TypeSpecifier
from enum import Enum, Flag

__all__ = ['StructureMember', 'StructureType', 'EnumType', 'FlagType', 'TupleType', 'StateTypeT', 'MoveTypeT', 'PhysicsTypeT', 'ColorType', 'ColorMultType', 'TransTypeT', 'AssertTypeT', 'FloatPairType', 'WaveTypeT', 'HelperTypeT', 'HitFlagTypeF', 'GuardFlagTypeF', 'TeamTypeT', 'HitAnimTypeT', 'AttackTypeT', 'PriorityTypeT', 'PosTypeT', 'FloatPosType', 'IntPairType', 'WaveTupleType', 'HitTypeF', 'HitAttrF', 'HitStringType', 'PriorityPairType', 'SoundPairType', 'PeriodicColorType', 'BoolPairType', 'SpaceTypeT']

@dataclass
class StructureMember:
    name: str
    type: TypeSpecifier

class StructureType(TypeSpecifier):
    members: list[StructureMember]
    name: str
    category: TypeCategory
    register: bool
    def __init__(self, name: str, members: list[StructureMember], register: bool = True, library: str | None = None) -> None: ...

class EnumType(TypeSpecifier):
    members: list[str]
    name: str
    category: TypeCategory
    register: bool
    library: str | None
    inner_type: type[Enum] | None
    def __init__(self, name: str, members: list[str] | type[Enum], register: bool = True, library: str | None = None) -> None: ...
    def __getattr__(self, name: str) -> Expression: ...

class FlagType(TypeSpecifier):
    members: list[str]
    name: str
    category: TypeCategory
    register: bool
    library: str | None
    inner_type: type[Flag] | None
    def __init__(self, name: str, members: list[str] | type[Flag], register: bool = True, library: str | None = None) -> None: ...
    def __getattr__(self, name: str) -> Expression: ...

class TupleType(TypeSpecifier):
    members: list[TypeSpecifier]
    name: str
    category: TypeCategory
    def __init__(self, name: str, category: TypeCategory, members: list[TypeSpecifier]) -> None: ...

StateTypeT: EnumType
MoveTypeT: EnumType
PhysicsTypeT: EnumType
HitTypeF: FlagType
HitAttrF: FlagType
TransTypeT: EnumType
AssertTypeT: EnumType
WaveTypeT: EnumType
HelperTypeT: EnumType
TeamTypeT: EnumType
HitAnimTypeT: EnumType
AttackTypeT: EnumType
PriorityTypeT: EnumType
PosTypeT: EnumType
SpaceTypeT: EnumType
TeamModeTypeT: EnumType
HitFlagTypeF: FlagType
GuardFlagTypeF: FlagType
ColorType: TupleType
ColorMultType: TupleType
FloatPairType: TupleType
FloatPosType: TupleType
IntPairType: TupleType
BoolPairType: TupleType
WaveTupleType: TupleType
HitStringType: TupleType
PriorityPairType: TupleType
SoundPairType: TupleType
PeriodicColorType: TupleType



from mdk.types.specifier import TypeSpecifier as TypeSpecifier

def check_types_assignable(spec1: TypeSpecifier, spec2: TypeSpecifier) -> TypeSpecifier | None: ...
def check_any_assignable(spec1: TypeSpecifier, spec2: list[TypeSpecifier]) -> TypeSpecifier | None: ...



## utilities for loading sprite data from SFF files.
## leaning heavily on the reference in https://github.com/bitcraft/mugen-tools/blob/master/libmugen/sff.py as SFF's implementation of some things is wonky
## thanks also to https://sourceforge.net/projects/nomeneditor/ for RLE5 and LZ5 implementations.
from __future__ import annotations

from io import BufferedReader
from struct import unpack
from dataclasses import dataclass

def readbyte(io: BufferedReader) -> int:
    return io.read(1)[0]

def readshort(io: BufferedReader) -> int:
    return unpack('<H', io.read(2))[0]

def readint(io: BufferedReader) -> int:
    return unpack('<I', io.read(4))[0]

@dataclass
class SFFHeader:
    signature: bytes
    verlo3: int
    verlo2: int
    verlo1: int
    verhi: int
    compatverlo3: int
    compatverlo2: int
    compatverlo1: int
    compatverhi: int
    sprite_offset: int
    sprite_total: int
    palette_offset: int
    palette_total: int
    ldata_offset: int
    ldata_total: int
    tdata_offset: int
    tdata_total: int

    @classmethod
    def load(cls, io: BufferedReader) -> SFFHeader:
        sig = io.read(12)
        vers = (readbyte(io), readbyte(io), readbyte(io), readbyte(io))
        _ = io.read(8)
        compat = (readbyte(io), readbyte(io), readbyte(io), readbyte(io))
        _ = io.read(8)
        sprite = (readint(io), readint(io))
        palette = (readint(io), readint(io))
        ldata = (readint(io), readint(io))
        tdata = (readint(io), readint(io))

        return SFFHeader(
            sig,
            vers[0], vers[1], vers[2], vers[3],
            compat[0], compat[1], compat[2], compat[3],
            sprite[0], sprite[1],
            palette[0], palette[1],
            ldata[0], ldata[1],
            tdata[0], tdata[1]
        )
    
@dataclass
class SFFPalette:
    groupno: int
    itemno: int
    numcols: int
    index: int
    data_offset: int
    data_length: int

    buffer: BufferedReader
    _cache: list[tuple[int, int, int, int]] | None = None

    @classmethod
    def load(cls, io: BufferedReader, header: SFFHeader, member: int) -> SFFPalette:
        # each palette header is 16 bytes long
        offs = 16 * member
        # seek to the offset for this member
        io.seek(header.palette_offset + offs)

        group = readshort(io)
        item = readshort(io)
        cols = readshort(io)
        index = readshort(io)
        data = (readint(io), readint(io))

        return SFFPalette(
            group, item, cols, index, 
            data[0] + header.ldata_offset, data[1], io
        )
    
    @property
    def data(self) -> list[tuple[int, int, int, int]]:
        if self._cache != None:
            return self._cache

        ## palette data sits in the ldata chunk
        result: list[tuple[int, int, int, int]] = []
        self.buffer.seek(self.data_offset)

        index = 0
        while self.buffer.tell() < self.data_offset + self.data_length:
            colors = (readbyte(self.buffer), readbyte(self.buffer), readbyte(self.buffer), 0 if index == 0 else 255)
            _ = readbyte(self.buffer)
            result.append(colors)
            index += 1

        self._cache = result
        return result

@dataclass
class SFFSprite:
    groupno: int
    itemno: int
    width: int
    height: int
    offsetx: int
    offsety: int
    index: int
    format: int
    depth: int
    data_offset: int
    data_length: int
    palette_index: int
    flags: int

    buffer: BufferedReader
    _cache: bytes | None = None

    @classmethod
    def load(cls, io: BufferedReader, header: SFFHeader, member: int) -> SFFSprite:
        # each sprite header is 28 bytes long
        offs = 28 * member
        # seek to the offset for this member
        io.seek(header.sprite_offset + offs)

        group = readshort(io)
        item = readshort(io)
        width = readshort(io)
        height = readshort(io)
        offset = (readshort(io), readshort(io))
        index = readshort(io)
        format = readbyte(io)
        depth = readbyte(io)
        data = (readint(io), readint(io))
        palette = readshort(io)
        flags = readshort(io)

        # determine whether to load from ldata or tdata
        if flags == 0:
            data = (data[0] + header.ldata_offset, data[1])
        else:
            data = (data[0] + header.tdata_offset, data[1])

        return SFFSprite(
            group, item, width, height,
            offset[0], offset[1], index, format,
            depth, data[0], data[1], palette, flags,
            io
        )
    
    @property
    def data(self) -> bytes:
        if self._cache != None:
            return self._cache

        ## sprite data is already offset to either ldata or tdata depending on flags
        self.buffer.seek(self.data_offset)
        size = readint(self.buffer)
        raw = self.buffer.read(self.data_length - 4)

        ## raw data may need to be decoded (e.g. rle8)
        if self.format == 0:
            self._cache = raw
        elif self.format == 1:
            raise Exception(f"Invalid format used for sprite {self.groupno, self.itemno}")
        elif self.format == 2:
            self._cache = self.rle8(raw, size)
        elif self.format == 3:
            self._cache = self.rle5(raw, size)
        elif self.format == 4:
            self._cache = self.lz5(raw, size)
        else:
            raise Exception(f"Don't currently know how to decode format {self.format} for sprite {self.groupno, self.itemno}")
        
        return self._cache
    
    def palette(self, pal: SFFPalette) -> bytes:
        ## this function expects the output from `self.data` to be palette-indexed colors.
        colors = pal.data
        indexes = self.data

        ## raw format can only be paletted if the color depth is 8, otherwise color info is encoded into the raw data.
        if self.format == 0 and self.depth != 8:
            raise Exception("Raw-format image cannot have a palette applied unless it uses color depth of 8.")
        
        ## get the pixel data into an array
        pixels = [color for pixel in indexes for color in colors[pixel]]
        
        return bytes(pixels)
        
    def rle8(self, raw: bytes, size: int) -> bytes:
        ## this function produces a block of bytes without paletting applied.
        result = bytearray()

        index = 0
        while index < len(raw):
            ch = raw[index]
            ## if ch & 0xC0 == 0x40, then (ch & 0x3F) encodes the runlength, and the next byte encodes the color.
            if (ch & 0xC0) == 0x40:
                index += 1
                color = raw[index]
                for _ in range(ch & 0x3F): result.append(color)
            else:
                result.append(ch)
            index += 1

        return bytes(result)

    def rle5(self, raw: bytes, size: int) -> bytes:
        ## this function produces a block of bytes without paletting applied.
        result = bytearray()

        index = 0
        while index < len(raw):
            run = raw[index]
            proc = raw[index + 1]
            index += 2

            switch = (proc & 0x80) / 0x80
            data = (proc & 0x7F)

            if switch == 1:
                color = raw[index]
                index += 1
            else:
                color = 0

            for _ in range(run): result.append(color)
            for _ in range(data):
                b = raw[index]
                index += 1
                color = b & 0x1F
                run = b >> 5
                for _ in range(run): result.append(color)

        return bytes(result)

    def lz5(self, raw: bytes, size: int) -> bytes:
        ## this function produces a block of bytes without paletting applied.
        result = bytearray()

        index = 0
        packcnt = 0
        recycle = 0
        while index < len(raw):
            ## read control packet
            control = raw[index]
            index += 1

            ## read 8 packets, following flags from the control packet
            for cnt in range(8):
                if index >= len(raw): break
                mask = 2 ** cnt
                use_lz = (control & mask) != 0
                if use_lz:
                    ## LZ encoded
                    byte1 = raw[index]
                    index += 1
                    if (byte1 & 0x3F) == 0:
                        ## long LZ
                        byte2 = raw[index]
                        byte3 = raw[index + 1]
                        index += 2
                        offset = ((byte1 & 0xC0) << 2) + (byte2 + 1)
                        length = byte3 + 3
                    else:
                        ## short LZ
                        length = (byte1 & 0x3F) + 1
                        packcnt += 1
                        recycle = (recycle << 2) + ((byte1 & 0xC0) >> 6)
                        if packcnt % 4 == 0:
                            offset = recycle + 1
                            recycle = 0
                        else:
                            byte2 = raw[index]
                            index += 1
                            offset = byte2 + 1
                    begin = len(result) - offset
                    for i in range(length):
                        result.append(result[begin + i])
                else:
                    ## RLE encoded
                    byte1 = raw[index]
                    index += 1
                    value = (byte1 & 0x1F)
                    count = (byte1 & 0xE0) >> 5
                    if count == 0:
                        ## long RLE
                        byte2 = raw[index]
                        index += 1
                        count = byte2 + 8
                    for _ in range(count):
                        result.append(value)

        return result

@dataclass
class SFF:
    header: SFFHeader
    palettes: list[SFFPalette]
    sprites: list[SFFSprite]
    buffer: BufferedReader

    @classmethod
    def load(cls, io: BufferedReader) -> SFF:
        header = SFFHeader.load(io)
        palettes: list[SFFPalette] = []
        sprites: list[SFFSprite] = []

        for index in range(header.palette_total):
            palettes.append(SFFPalette.load(io, header, index))

        for index in range(header.sprite_total):
            sprites.append(SFFSprite.load(io, header, index))

        return SFF(
            header, 
            palettes, 
            sprites,
            io
        )
    
    def find_sprite(self, group: int, index: int) -> SFFSprite | None:
        for sprite in self.sprites:
            if sprite.groupno == group and sprite.itemno == index:
                ## handle linked sprites
                if sprite.data_length == 0:
                    return self.sprites[sprite.index]
                return sprite
        return None
    
    def find_palette(self, group: int, index: int) -> SFFPalette | None:
        for palette in self.palettes:
            if palette.groupno == group and palette.itemno == index:
                return palette
        return None


import traceback
import random
import string
import sys
from typing import Union, Callable
from functools import partial
from enum import Enum, Flag

from mdk.types.errors import CompilationException
from mdk.types.context import CompilerContext, StateController
from mdk.types.expressions import Expression, TupleExpression
from mdk.types.builtins import BoolType, IntType, FloatType, StringType, StateNoType, AnyType
from mdk.types.defined import TupleType

def convert(input: Union[Expression, Enum, Flag, str, int, float, bool, tuple, Callable[..., StateController]]) -> Expression:
    if isinstance(input, Expression):
        return input
    elif isinstance(input, partial):
        return Expression(input.keywords["value"], StateNoType)
    elif isinstance(input, Callable):
        return Expression(input.__name__, StateNoType)
    elif isinstance(input, tuple):
        elements: list[str] = []
        for elem in input:
            elements.append(convert(elem).exprn)
        return Expression(", ".join(elements), AnyType)
    elif type(input) == str:
        return Expression(input, StringType)
    elif type(input) == int:
        return Expression(str(input), IntType)
    elif type(input) == float:
        return Expression(str(input), FloatType)
    elif type(input) == bool:
        return Expression("true" if input else "false", BoolType)
    elif isinstance(input, Flag):
        ctx = CompilerContext.instance()
        for typename in ctx.typedefs:
            typedef = ctx.typedefs[typename]
            if hasattr(typedef, "inner_type") and typedef.inner_type == type(input): # type: ignore
                if not typedef.register:
                    result = ""
                    for member in input:
                        result += member.name # type: ignore
                    return Expression(f"{typedef.name}.{result}", typedef)
                else:
                    result = " | ".join([x.__str__() for x in input])
                    return Expression(f"({result})", typedef)
        raise Exception(f"Could not determine the MTL type to use for flag type {type(input)}.")
    elif isinstance(input, Enum):
        ctx = CompilerContext.instance()
        for typename in ctx.typedefs:
            typedef = ctx.typedefs[typename]
            if hasattr(typedef, "inner_type") and typedef.inner_type == type(input): # type: ignore
                return Expression(f"{typedef.name}.{input.name}", typedef)
        raise Exception(f"Could not determine the MTL type to use for enum type {type(input)}.")
    else:
        raise Exception(f"Attempted to convert from unsupported builtin type {type(input)}.")

def generate_random_string(length: int):
    return ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(length))

def format_tuple(t: tuple) -> TupleExpression:
    result: list[Expression] = []

    for index in range(len(t)):
        if isinstance(t[index], Expression):
            result.append(t[index])
        else:
            result.append(convert(t[index]))

    return tuple(result)

def convert_tuple(t: tuple, t2: TupleType) -> Expression:
    expression = format_tuple(t)
    return Expression(", ".join([t.exprn for t in expression]), t2) # type: ignore

def format_bool(b: bool) -> Expression:
    return Expression("true" if b else "false", BoolType)

def create_compiler_error(exc: CompilationException):
    context = CompilerContext.instance()
    ## extract the portion of the stack trace that is actually relevant...
    _exc = exc
    if exc.__context__ != None:
        _exc = exc.__context__
    tb = traceback.extract_tb(_exc.__traceback__)
    ## we want to identify the user-side issue (because the traceback contains a bunch of MDK internals as well)
    save_lines: list[str] = []
    for fs in tb:
        for tm in context.templates:
            if context.templates[tm].fn.__name__ == fs.name: save_lines.append(f"{fs.filename}:{fs.lineno}\n\t{fs.line}")
        for sd in context.statedefs:
            if context.statedefs[sd].fn.__name__ == fs.name: save_lines.append(f"{fs.filename}:{fs.lineno}\n\t{fs.line}")
        for fn in context.statefuncs:
            if fn == fs.name: save_lines.append(f"{fs.filename}:{fs.lineno}\n\t{fs.line}")
    ## now print full exception and likely causes.
    traceback.print_exception(_exc)
    print()
    print("Likely cause(s) in user-code at:")
    print("\n".join(save_lines))
    print()
    sys.exit(-1)


from typing import Callable, NoReturn, Literal, Protocol, Optional

from mdk.types.context import StateScope
from mdk.types.expressions import ConvertibleExpression, Expression
from mdk.types.specifier import TypeSpecifier

__all__ = ['VariableExpression', 'IntVar', 'FloatVar', 'ShortVar', 'ByteVar', 'BoolVar']

class VariableExpression(Expression):
    type: TypeSpecifier
    exprn: str
    def __init__(self, type: TypeSpecifier, scope: StateScope | None = None, system: bool = False) -> None: ...
    def make_expression(self, exprn: str): ...
    def set(self, val: ConvertibleExpression) -> None: ...
    def add(self, val: ConvertibleExpression) -> None: ...

    def __eq__(self, other: ConvertibleExpression) -> Expression: ... # type: ignore
    def __ne__(self, other: ConvertibleExpression) -> Expression: ... # type: ignore
    def __lt__(self, other: ConvertibleExpression) -> Expression: ...
    def __le__(self, other: ConvertibleExpression) -> Expression: ...
    def __gt__(self, other: ConvertibleExpression) -> Expression: ...
    def __ge__(self, other: ConvertibleExpression) -> Expression: ...
    def __add__(self, other: ConvertibleExpression) -> Expression: ...
    def __sub__(self, other: ConvertibleExpression) -> Expression: ...
    def __mul__(self, other: ConvertibleExpression) -> Expression: ...
    def __truediv__(self, other: ConvertibleExpression) -> Expression: ...
    def __floordiv__(self, other: ConvertibleExpression) -> Expression: ...
    def __mod__(self, other: ConvertibleExpression) -> Expression: ...
    def __pow__(self, other: ConvertibleExpression) -> Expression: ...
    def __rshift__(self, other: ConvertibleExpression) -> NoReturn: ...
    def __lshift__(self, other: ConvertibleExpression) -> NoReturn: ...
    def __and__(self, other: ConvertibleExpression) -> Expression: ...
    def __or__(self, other: ConvertibleExpression) -> Expression: ...
    def __xor__(self, other: ConvertibleExpression) -> Expression: ...
    def __radd__(self, other: ConvertibleExpression) -> Expression: ...
    def __rsub__(self, other: ConvertibleExpression) -> Expression: ...
    def __rmul__(self, other: ConvertibleExpression) -> Expression: ...
    def __rtruediv__(self, other: ConvertibleExpression) -> Expression: ...
    def __rfloordiv__(self, other: ConvertibleExpression) -> Expression: ...
    def __rmod__(self, other: ConvertibleExpression) -> Expression: ...
    def __rpow__(self, other: ConvertibleExpression) -> Expression: ...
    def __rlshift__(self, other: ConvertibleExpression) -> NoReturn: ...
    def __rrshift__(self, other: ConvertibleExpression) -> NoReturn: ...
    def __rand__(self, other: ConvertibleExpression) -> Expression: ...
    def __ror__(self, other: ConvertibleExpression) -> Expression: ...
    def __rxor__(self, other: ConvertibleExpression) -> Expression: ...
    def __neg__(self) -> Expression: ...
    def __pos__(self) -> Expression: ...
    def __abs__(self) -> Expression: ...
    def __invert__(self) -> Expression: ...
    def __round__(self) -> Expression: ...
    def __trunc__(self) -> Expression: ...
    def __floor__(self) -> Expression: ...
    def __ceil__(self) -> Expression: ...
    def __bool__(self) -> Literal[True]: ...

class VariableType(Protocol):
    def __call__(self, /, scope: Optional[StateScope] = ..., system: bool = ...) -> VariableExpression:
        ...

IntVar: VariableType
ShortVar: VariableType
ByteVar: VariableType
BoolVar: VariableType
FloatVar: VariableType



from mdk.resources.animation import Animation, Sequence, Frame, AnimationFlip, Clsn

def read_animations(anim_path: str) -> list[Animation]:
    animations: list[Animation] = []
    with open(anim_path) as f:
        content = f.readlines()
        current_anim: Animation | None = None
        next_loop: bool = False
        clsns: list[dict] = []
        clsns_default: bool = False
        for line in content:
            # skip/eliminate comments
            if line.strip().startswith(";"):
                continue
            line = line.strip().split(';')[0]
            
            if line.strip().lower().startswith("[begin action "):
                target_index = line.strip().lower()[14:].split(']')[0].strip()
                print(f"Discovered new external animation with ID {target_index}")
                if current_anim != None and current_anim._id != None:
                    animations.append(current_anim)
                current_anim = Animation(frames = [], id = int(target_index), external = False)
                next_loop = False
                clsns_default = False
                clsns = []
            elif line.strip().lower() == "loopstart":
                next_loop = True
            elif line.strip().lower().startswith("clsn1default:") or line.strip().lower().startswith("clsn2default:"):
                ## mark upcoming CLSNs as default
                clsns_default = True
            elif line.strip().lower().startswith("clsn1:") or line.strip().lower().startswith("clsn2:"):
                ## mark upcoming CLSNs as non-default
                clsns_default = False
            elif line.strip().lower().startswith("clsn1["):
                ## read clsn
                data = [int(x.strip()) for x in line.strip().split("=")[1].split(",")]
                clsns.append({
                    "data": data,
                    "default": clsns_default,
                    "type": 1
                })
            elif line.strip().lower().startswith("clsn2["):
                ## read clsn
                data = [int(x.strip()) for x in line.strip().split("=")[1].split(",")]
                clsns.append({
                    "data": data,
                    "default": clsns_default,
                    "type": 2
                })
            elif "," in line and current_anim != None and current_anim._frames != None:
                params = [s.strip().lower() for s in line.split(",")]
                if len(params) < 5: continue
                next_frame = Frame(int(params[0]), int(params[1]))
                next_frame.offset((int(params[2]), int(params[3])))
                next_frame.length(int(params[4]))
                if len(params) >= 6 and params[5] != "":
                    flip = AnimationFlip.NONE
                    if params[5] == "H": flip = AnimationFlip.HORIZONTAL
                    if params[5] == "V": flip = AnimationFlip.VERTICAL
                    if params[5] == "VH" or params[5] == "HV": flip = AnimationFlip.BOTH
                    if flip != AnimationFlip.NONE: next_frame.flip(flip)
                if len(params) >= 7 and params[6] != "":
                    next_frame._trans = params[6]
                if len(params) >= 9 and params[7] != "" and params[8] != "":
                    next_frame.scale((float(params[7]), float(params[8])))
                if len(params) >= 10 and params[9] != "":
                    next_frame.rotate(int(params[9]))
                if next_loop:
                    next_frame.loop()
                    next_loop = False
                for clsn in clsns:
                    if clsn['type'] == 1:
                        next_frame.clsn1(Clsn(clsn['data'][0], clsn['data'][1], clsn['data'][2], clsn['data'][3], clsn['default']))
                    elif clsn['type'] == 2:
                        next_frame.clsn2(Clsn(clsn['data'][0], clsn['data'][1], clsn['data'][2], clsn['data'][3], clsn['default']))
                    else:
                        raise Exception(f"Unrecognized external clsn type {clsn['type']}")
                current_anim._frames._frames.append(next_frame)
                clsns_default = False
                clsns = []
        if current_anim != None and current_anim._id != None:
            animations.append(current_anim)
    return animations


from typing import Callable
from enum import Enum, Flag

from mdk.types.builtins import BoolType as BoolType, FloatType as FloatType, IntType as IntType, StringType as StringType, StateNoType as StateNoType
from mdk.types.context import CompilerContext as CompilerContext, StateController
from mdk.types.defined import TupleType as TupleType
from mdk.types.errors import CompilationException as CompilationException
from mdk.types.expressions import Expression as Expression, TupleExpression as TupleExpression

def convert(input: Expression | str | int | float | bool | tuple | Callable[..., StateController] | Enum | Flag) -> Expression: ...
def generate_random_string(length: int): ...
def format_tuple(t: tuple) -> TupleExpression: ...
def convert_tuple(t: tuple, t2: TupleType) -> Expression: ...
def format_bool(b: bool) -> Expression: ...
def create_compiler_error(exc: CompilationException): ...



from typing import Callable, Optional, Union
from enum import Enum, Flag
import copy
import functools
import traceback

from mdk.types.context import StateController, CompilerContext
from mdk.types.expressions import Expression, TupleExpression
from mdk.types.builtins import IntType, StateNoType, AnyType, BoolType
from mdk.types.defined import TupleType
from mdk.types.specifier import TypeSpecifier

from mdk.utils.shared import format_bool, format_tuple, convert
from mdk.utils.expressions import check_any_assignable

from mdk.resources.animation import Animation

# decorator which provides a wrapper around each controller.
# this adds some extra debugging info, and also simplifies adding triggers to controllers and handling controller insertion into the active statedef.
def controller(**typeinfo) -> Callable[[Callable[..., StateController]], Callable[..., StateController]]:
    def wrapper(fn: Callable[..., StateController]) -> Callable[..., StateController]:
        def decorated(*args, ignorehitpause: Optional[bool] = None, persistent: Optional[int] = None, **kwargs) -> StateController:
            return make_controller(fn, *args, typeinfo = typeinfo, ignorehitpause = ignorehitpause, persistent = persistent, **kwargs)
        return decorated
    return wrapper
    
def make_controller(fn, *args, typeinfo: dict[str, list[Optional[TypeSpecifier]]], ignorehitpause: Optional[bool] = None, persistent: Optional[int] = None, **kwargs) -> StateController:
    new_kwargs: dict[str, Union[Expression, TupleExpression]] = {}
    for name in typeinfo:
        valid_types = typeinfo[name]
        valid_no_none = [t for t in valid_types if t != None]
        valid_typenames = [type.name for type in valid_types if type != None]
        ## 1. ensure required params are included
        if None not in valid_types and name not in kwargs:
            raise Exception(f"Controller {fn.__name__} has a required parameter {name} which was not provided.")
        ## 2. ensure types are correct
        if name in kwargs:
            input_expression = kwargs[name]
            if type(input_expression) in [int, str, float, bool]:
                input_expression = convert(input_expression)
            elif isinstance(input_expression, Flag) or isinstance(input_expression, Enum):
                input_expression = convert(input_expression)
            elif isinstance(input_expression, Animation):
                if input_expression._id == None:
                    raise Exception("Animation without an assigned ID cannot be used as a Controller parameter! (Did you try to use an automatic-numbered Animation in global state?)")
                input_expression = convert(input_expression._id)
            if isinstance(input_expression, Expression):
                input_type = input_expression.type
                if check_any_assignable(input_type, valid_no_none) == None:
                    raise Exception(f"Parameter {name} on controller {fn.__name__} expects an expression with a type from ({', '.join(valid_typenames)}), but got {input_type.name} instead.")
                new_kwargs[name] = input_expression
            elif type(input_expression) == tuple:
                ## for a tuple, we expect TupleExpression in the valid types. there should be exactly 1 valid type.
                if len(valid_no_none) != 1 or (not isinstance(valid_no_none[0], TupleType) and valid_no_none[0] != AnyType):
                    raise Exception(f"Controller {fn.__name__} has a parameter {name} which expects an expression with a type from ({', '.join(valid_typenames)}), but parameter required {type(valid_no_none[0])} - bug the developers.")
                target_type = valid_no_none[0]
                ## although we provide typings for the tuple, we don't actually check them here.
                ## (mdk-python's current API does not encode optional/repeated tuple members, so the type check is incorrect anyway)
                ## this is fine since MTL will catch any mistakes during translation, as MTL itself supports these features.
                new_kwargs[name] = format_tuple(input_expression)
            elif type(input_expression) in [functools.partial, Callable] and StateNoType in valid_types:
                ## you're allowed to use function pointers directly as stateno types.
                new_kwargs[name] = input_expression
            elif input_expression != None:
                ## required parameters were already checked, so if None was passed explicitly it's completely fine. we just skip that parameter.
                raise Exception(f"Couldn't determine input type for parameter {name} on controller {fn.__name__}; input type was {type(input_expression)}.")

    ctrl: StateController = fn(*args, **new_kwargs)
    ctrl.type = fn.__name__

    ctx = CompilerContext.instance()
    if ctx.default_state[0] != None: ctrl.params["ignorehitpause"] = ctx.default_state[0]
    if ctx.default_state[1] != None: ctrl.params["persistent"] = ctx.default_state[1]

    if ignorehitpause != None: ctrl.params["ignorehitpause"] = format_bool(ignorehitpause)
    if persistent != None: ctrl.params["persistent"] = Expression(str(persistent), IntType)
    
    if ctx.current_state == None and ctx.current_template == None:
        raise Exception("Attempted to call a state controller outside of a statedef function.")
    
    ctrl.triggers = copy.deepcopy(ctx.trigger_stack)
    if len(ctx.if_stack) != 0:
        ctrl.triggers.append(Expression(f"mdk_internalTrigger = {ctx.if_stack[-1]}", BoolType))
    if len(ctx.check_stack) != 0:
        ctrl.triggers = copy.deepcopy(ctx.check_stack) + ctrl.triggers
        ctx.check_stack = []
    ctx.current_trigger = None

    ## i believe this should always be right, since it flows from <callsite> -> decorated -> make_controller
    callsite = traceback.extract_stack()[-3]
    ctrl.location = (callsite.filename, callsite.lineno if callsite.lineno != None else 0)
    
    if ctx.current_state != None:
        ctx.current_state.controllers.append(ctrl)
    elif ctx.current_template != None:
        ctx.current_template.controllers.append(ctrl)
    return ctrl


import io
import inspect
import ast
import types
from typing import Callable

from mdk.types.context import StateController, TranslationMode

from mdk.utils.shared import format_bool
from mdk.utils.triggers import TriggerAnd, TriggerOr, TriggerNot, TriggerNotIn, TriggerAssign, TriggerPush, TriggerPop, TriggerPrint

def write_controller(ctrl: StateController, f: io.TextIOWrapper, locations: bool):
    f.write("[State ]\n")
    f.write(f"type = {ctrl.type}\n")
    if len(ctrl.triggers) == 0: ctrl.triggers.append(format_bool(True))
    for trigger in ctrl.triggers:
        f.write(f"trigger1 = {trigger.__str__()}\n")
    for param in ctrl.params:
        f.write(f"{param} = {ctrl.params[param].__str__()}\n")
    if ctrl.location[0] != "<?>" and ctrl.location[1] != 0 and locations:
        f.write(f"mtl.location.file = {ctrl.location[0]}\n")
        f.write(f"mtl.location.line = {ctrl.location[1]}\n")

def rewrite_function(fn: Callable[..., None], overload_print: bool = True, mode: TranslationMode = TranslationMode.STANDARD) -> Callable[..., None]:
    # get effective source code of the decorated function
    source, line_number = inspect.getsourcelines(fn)
    location = inspect.getsourcefile(fn)
    # remove decorator lines at the start of the source
    while source[0].strip().startswith('@'):
        source = source[1:]
    source = ''.join(source)
    # parse AST from the decorated function
    old_ast = ast.parse(source)
    # use a node transformer to replace any operators we can't override behaviour of (e.g. `and`, `or`, `not`) with function calls
    new_ast = ReplaceLogicalOperators(overload_print, mode = mode).visit(old_ast)
    ast.increment_lineno(new_ast, line_number)
    # compile the updated AST to a function and use it as the resulting wrapped function.
    new_code_obj = compile(new_ast, fn.__code__.co_filename, 'exec')
    # add missing globals for each operation.
    # these globals are placed into `mdk.impl` namespace to make it easier to identify them in error cases.
    new_globals = fn.__globals__
    new_globals["mdk.impl.TriggerAnd"] = TriggerAnd
    new_globals["mdk.impl.TriggerOr"] = TriggerOr
    new_globals["mdk.impl.TriggerNot"] = TriggerNot
    new_globals["mdk.impl.TriggerNotIn"] = TriggerNotIn
    new_globals["mdk.impl.TriggerAssign"] = TriggerAssign
    new_globals["mdk.impl.TriggerPush"] = TriggerPush
    new_globals["mdk.impl.TriggerPop"] = TriggerPop
    new_globals["mdk.impl.TriggerPrint"] = TriggerPrint
    ## find the last code object in the const list.
    new_fn = None
    for index in range(len(new_code_obj.co_consts)):
        if type(new_code_obj.co_consts[index]) == types.CodeType:
            new_fn = types.FunctionType(new_code_obj.co_consts[index], new_globals)
    if new_fn == None:
        raise Exception("Failed to find function code object during template patchup.")
    
    return new_fn

class ReplaceLogicalOperators(ast.NodeTransformer):
    def __init__(self, prints: bool, mode: TranslationMode):
        self.prints = prints
        self.if_depth = 0
        self.mode = mode
        super().__init__()

    def visit_Call(self, node: ast.Call):
        ## if this Call is a print, we want to override with TriggerPrint.
        if not self.prints: return node

        if isinstance(node.func, ast.Name) and node.func.id == 'print':
            node.func.id = "mdk.impl.TriggerPrint"

        return node

    def visit_BoolOp(self, node: ast.BoolOp):
        # inspect child nodes.
        node = super(ReplaceLogicalOperators, self).generic_visit(node) # type: ignore
        # get the type of operation: we support transforming AND, OR for boolop.
        if type(node.op) == ast.And:
            target = 'mdk.impl.TriggerAnd'
        elif type(node.op) == ast.Or:
            target = 'mdk.impl.TriggerOr'
        else:
            return node
        
        # then, replace the BoolOp directly with a Call to the appropriate override,
        # with arguments provided from the inner values.
        return ast.Call(
            func=ast.Name(id=target, ctx=ast.Load(), lineno=node.lineno, col_offset=node.col_offset),
            args=node.values,
            keywords=[],
            lineno=node.lineno,
            col_offset=node.col_offset
        )
    
    def visit_Compare(self, node: ast.Compare):
        # recursively inspect child nodes.
        node = super(ReplaceLogicalOperators, self).generic_visit(node) # type: ignore
        # get the type of operation: we support transforming NOT-IN for cmpop (for Flag membership behaviour)
        if len(node.ops) == 1 and type(node.ops[0]) == ast.NotIn:
            print(ast.dump(node, indent=4))
            target = 'mdk.impl.TriggerNotIn'
        else:
            return node
        
        # then, replace the CmpOp directly with a Call to the appropriate override,
        # with arguments provided from the inner values.
        return ast.Call(
            func=ast.Name(id=target, ctx=ast.Load(), lineno=node.lineno, col_offset=node.col_offset),
            args=[
                node.left,
                node.comparators[0]
            ],
            keywords=[],
            lineno=node.lineno,
            col_offset=node.col_offset
        )
    
    def visit_UnaryOp(self, node: ast.UnaryOp):
        # recursively inspect child nodes.
        node = super(ReplaceLogicalOperators, self).generic_visit(node) # type: ignore
        # get the type of operation: we support transforming NOT for unaryop
        if type(node.op) == ast.Not:
            target = 'mdk.impl.TriggerNot'
        else:
            return node
        
        # then, replace the BoolOp directly with a Call to the appropriate override,
        # with arguments provided from the inner values.
        return ast.Call(
            func=ast.Name(id=target, ctx=ast.Load(), lineno=node.lineno, col_offset=node.col_offset),
            args=[
                node.operand
            ],
            keywords=[],
            lineno=node.lineno,
            col_offset=node.col_offset
        )
    
    def visit_NamedExpr(self, node: ast.NamedExpr):
        # recursively inspect child nodes.
        node = super(ReplaceLogicalOperators, self).generic_visit(node) # type: ignore
        # then, replace the NamedExpr directly with a Call to the assignment override,
        # with arguments provided from the inner values.
        return ast.Call(
            func=ast.Name(id="mdk.impl.TriggerAssign", ctx=ast.Load(), lineno=node.lineno, col_offset=0),
            args=[
                ast.Name(id=node.target.id, ctx=ast.Load(), lineno=node.lineno, col_offset=0),
                node.value
            ],
            keywords=[],
            lineno=node.lineno,
            col_offset=node.col_offset
        )
    
    def doIf_Standard(self, node: ast.If):
        # recursively inspect child nodes.
        node = super(ReplaceLogicalOperators, self).generic_visit(node) # type: ignore

        results: list[ast.stmt] = []

        # append a call to mdk.impl.TriggerPush at the start of the block,
        # and append a call to mdk.impl.TriggerPop at the end of the block.
        node.body.insert(0, ast.Expr(
            value=ast.Call(
                func=ast.Name(id="mdk.impl.TriggerPush", ctx=ast.Load(), lineno=node.lineno, col_offset=0),
                args=[
                    ast.Constant(value=None, lineno=node.lineno, col_offset=0)
                ],
                keywords=[],
                lineno=node.lineno,
                col_offset=node.col_offset
            ),
            lineno=node.lineno,
            col_offset=node.col_offset
        ))
        node.body.insert(len(node.body), ast.Expr(
            value=ast.Call(
                func=ast.Name(id="mdk.impl.TriggerPop", ctx=ast.Load(), lineno=node.lineno, col_offset=0),
                args=[
                    ast.Constant(value=None, lineno=node.lineno, col_offset=0)
                ],
                keywords=[],
                lineno=node.lineno,
                col_offset=node.col_offset
            ),
            lineno=node.lineno,
            col_offset=node.col_offset
        ))

        results.append(node)

        # if the 'orelse' block does not start with an 'if' (i.e. is not an 'elif')
        if len(node.orelse) == 1 and isinstance(node.orelse, ast.If):
            ## this is the case for 'elif' blocks.
            ## just pull the If out of the orelse and append it.
            results.append(node.orelse)
            node.orelse = []
        elif len(node.orelse) != 0:
            ## this is the case for 'else' blocks.
            ## we need to wrap these up in another If with the negation of the previous statement.
            new_node = ast.If(
                test=ast.Call(
                    func=ast.Name(id='mdk.impl.TriggerNot', ctx=ast.Load(), lineno=node.lineno, col_offset=0),
                    args=[
                        node.test
                    ],
                    keywords=[],
                    lineno=node.lineno,
                    col_offset=node.col_offset
                ),
                body=node.orelse,
                orelse=[],
                lineno=node.lineno,
                col_offset=node.col_offset
            )
            new_node.body.insert(0, ast.Expr(
                value=ast.Call(
                    func=ast.Name(id="mdk.impl.TriggerPush", ctx=ast.Load(), lineno=node.lineno, col_offset=0),
                    args=[
                        ast.Constant(value=None, lineno=node.lineno, col_offset=0)
                    ],
                    keywords=[],
                    lineno=node.lineno,
                    col_offset=node.col_offset
                ),
                lineno=node.lineno,
                col_offset=node.col_offset
            ))
            new_node.body.insert(len(new_node.body), ast.Expr(
                value=ast.Call(
                    func=ast.Name(id="mdk.impl.TriggerPop", ctx=ast.Load(), lineno=node.lineno, col_offset=0),
                    args=[
                        ast.Constant(value=None, lineno=node.lineno, col_offset=0)
                    ],
                    keywords=[],
                    lineno=node.lineno,
                    col_offset=node.col_offset
                ),
                lineno=node.lineno,
                col_offset=node.col_offset
            ))
            results.append(new_node)
            node.orelse = []

        return results
    
    def doIf_Variable(self, node: ast.If):
        self.if_depth += 1
        results: list[ast.stmt] = []
        # recursively inspect child nodes.
        node = super(ReplaceLogicalOperators, self).generic_visit(node) # type: ignore

        # append a call to mdk.impl.TriggerPush at the start of the block,
        # and append a call to mdk.impl.TriggerPop at the end of the block.
        node.body.insert(0, ast.Expr(
            value=ast.Call(
                func=ast.Name(id="mdk.impl.TriggerPush", ctx=ast.Load(), lineno=node.lineno, col_offset=0),
                args=[
                    ast.Constant(value=self.if_depth, lineno=node.lineno, col_offset=0)
                ],
                keywords=[],
                lineno=node.lineno,
                col_offset=node.col_offset
            ),
            lineno=node.lineno,
            col_offset=node.col_offset
        ))
        node.body.insert(len(node.body), ast.Expr(
            value=ast.Call(
                func=ast.Name(id="mdk.impl.TriggerPop", ctx=ast.Load(), lineno=node.lineno, col_offset=0),
                args=[
                    ast.Constant(value=self.if_depth, lineno=node.lineno, col_offset=0)
                ],
                keywords=[],
                lineno=node.lineno,
                col_offset=node.col_offset
            ),
            lineno=node.lineno,
            col_offset=node.col_offset
        ))

        results.append(node)

        ## flatten tree
        ### TODO: `node.orelse` contains either ast.If or other statements.
        ### other statements need to be bundled and dropped into an ast.If as well.
        else_stmt = []
        for stmt in node.orelse:
            if isinstance(stmt, ast.If):
                results.append(stmt)
            else:
                else_stmt.append(stmt)

        if len(else_stmt) > 0:
            results.append(ast.Expr(
                value=ast.Call(
                    func=ast.Name(id="mdk.impl.TriggerPush", ctx=ast.Load(), lineno=node.lineno, col_offset=0),
                    args=[
                        ast.Constant(value=-1, lineno=node.lineno, col_offset=0)
                    ],
                    keywords=[],
                    lineno=node.lineno,
                    col_offset=node.col_offset
                ),
                lineno=node.lineno,
                col_offset=node.col_offset
            ))

            results += else_stmt

            results.append(ast.Expr(
                value=ast.Call(
                    func=ast.Name(id="mdk.impl.TriggerPop", ctx=ast.Load(), lineno=node.lineno, col_offset=0),
                    args=[
                        ast.Constant(value=None, lineno=node.lineno, col_offset=0)
                    ],
                    keywords=[],
                    lineno=node.lineno,
                    col_offset=node.col_offset
                ),
                lineno=node.lineno,
                col_offset=node.col_offset
            ))

        node.orelse = []

        self.if_depth -= 1
        return results
    
    def visit_If(self, node: ast.If):
        ## if the input function is scoped to TARGET, or is a library (template/trigger) function,
        ## use the safe (non-variable-based) conditional expansion.
        if self.mode == TranslationMode.VARIABLE:
            return self.doIf_Variable(node)
        elif self.mode == TranslationMode.STANDARD:
            return self.doIf_Standard(node)

        


from dataclasses import dataclass
from enum import Enum
from mdk.types.expressions import Expression as Expression
from mdk.types.specifier import TypeSpecifier as TypeSpecifier
from typing import Callable, Optional

class TranslationMode(Enum):
    STANDARD = 0
    VARIABLE = 1

class StateScopeType(Enum):
    SHARED = 0
    PLAYER = 1
    HELPER = 2
    TARGET = 3

@dataclass
class StateScope:
    scope: StateScopeType
    target: Optional[int] = None

SCOPE_TARGET: StateScope
SCOPE_PLAYER: StateScope
SCOPE_HELPER: Callable[[Optional[int]], StateScope]

@dataclass
class StateController:
    type: str
    params: dict[str, Expression]
    triggers: list[Expression]
    location: tuple[str, int]
    def __init__(self) -> None: ...

@dataclass
class ParameterDefinition:
    type: TypeSpecifier
    name: str
    scope: StateScope | None = None
    is_system: bool = False

@dataclass
class StateDefinition:
    fn: Callable[[], None]
    params: dict[str, Expression]
    controllers: list[StateController]
    locals: list[ParameterDefinition]

@dataclass
class TemplateDefinition:
    fn: Callable[..., None]
    library: str | None
    params: dict[str, TypeSpecifier]
    controllers: list[StateController]
    locals: list[ParameterDefinition]

@dataclass
class TriggerDefinition:
    fn: Callable[..., Expression]
    library: str | None
    result: TypeSpecifier
    params: dict[str, TypeSpecifier]
    exprn: Expression | None = ...

@dataclass
class CompilerContext:
    statedefs: dict[str, StateDefinition]
    templates: dict[str, TemplateDefinition]
    triggers: dict[str, TriggerDefinition]
    typedefs: dict[str, TypeSpecifier]
    current_state: StateDefinition | None
    current_template: TemplateDefinition | None
    current_trigger: Expression | None
    trigger_stack: list[Expression]
    globals: list[ParameterDefinition]
    default_state: tuple[Expression | None, Expression | None]
    format_params: list[Expression]
    if_stack: list[int]
    def __init__(self) -> None: ...
    @classmethod
    def instance(cls) -> CompilerContext: ...



from enum import Enum
from typing import Optional

## this specifies a subset of the type categories provided by MTL.
class TypeCategory(Enum):
    #INVALID = -1
    #ALIAS = 0
    #UNION = 1
    ENUM = 2
    FLAG = 3
    STRUCTURE = 4
    TUPLE = 20
    #BUILTIN_STRUCTURE = 96 Note MDK does not need to differentiate between builtin and user-defined structs.
    #STRING_FLAG = 97 Note MDK does not need string flag/enum as it uses MTL for intermediate (where all enum/flag can be passed as string)
    #STRING_ENUM = 98
    BUILTIN = 99
    #BUILTIN_DENY = 100 Note MDK does not need to worry about BUILTIN_DENY as the MTL side can handle denying creation anyway.

## defines a generic type. this type is not intended for end-users, instead they should use a
## type specifier creation function (for user-defined enums, flags, and structs).
class TypeSpecifier:
    name: str
    category: TypeCategory
    library: Optional[str]
    register: bool
    def __init__(self, name: str, category: TypeCategory, register: bool = True):
        self.name = name
        self.category = category
        self.library = None
        self.register = register


from enum import Flag, Enum
from typing import Callable

class AnimationFlip(Flag):
    NONE = 0
    VERTICAL = 1
    HORIZONTAL = 2
    BOTH = 3

class AnimationTrans(Enum):
    NONE = 0
    ADD = 1
    SUB = 2

class Clsn:
    _xmin: int
    _ymin: int
    _xmax: int
    _ymax: int
    _default: bool

    def __init__(self, xmin: int, ymin: int, xmax: int, ymax: int, default: bool = False): ...

class Frame:
    _group: int
    _index: int
    _length: int = -1
    _offset: tuple[int, int] = (0, 0)
    _flip: AnimationFlip = AnimationFlip.NONE
    _trans: AnimationTrans = AnimationTrans.NONE
    _scale: tuple[float, float] = (1, 1)
    _rotate: int = 0
    _loopstart: bool = False

    def __init__(self, 
                 group: int, index: int, length: int = -1, 
                 offset: tuple[int, int] = (0, 0), flip: AnimationFlip = AnimationFlip.NONE, trans: AnimationTrans = AnimationTrans.NONE,
                 scale: tuple[float, float] = (1, 1), rotate: int = 0):
        ...

    def compile(self) -> str: ...

    def clsn1(self, clsn: Clsn) -> Frame: ...
    def clsn2(self, clsn: Clsn) -> Frame: ...
    def default(self) -> Frame: ...

    def seq(self) -> Sequence: ...
    def dup(self) -> Frame: ...
    def length(self, length: int) -> Frame: ...
    def flip(self, flip: AnimationFlip) -> Frame: ...
    def rotate(self, rotate: int) -> Frame: ...
    def offset(self, offset: tuple[int, int]) -> Frame: ...
    def translate(self, offset: tuple[int, int]) -> Frame: ...
    def scale(self, scale: tuple[float, float]) -> Frame: ...
    def loop(self) -> Frame: ...

class SequenceModifier:
    def seq(self) -> Sequence: ...
    def set(self, val: int | Enum, filter: Callable[[Frame], bool] | None = None) -> SequenceModifier: ...
    def add(self, val: int | Enum, filter: Callable[[Frame], bool] | None = None) -> SequenceModifier: ...
    def mul(self, val: int | Enum, filter: Callable[[Frame], bool] | None = None) -> SequenceModifier: ...
    def transform(self, transformer: Callable[[Frame, int], int]) -> SequenceModifier: ...
    def extend(self, frames: Frame | list[Frame] | Sequence | SequenceModifier | FrameSequenceModifier | TupleSequenceModifier) -> SequenceModifier: ...
    def __getitem__(self, key: int | slice) -> SequenceModifier: ...

    @property
    def frames(self) -> FrameSequenceModifier: ...
    @property
    def group(self) -> SequenceModifier: ...
    @property
    def index(self) -> SequenceModifier: ...
    @property
    def length(self) -> SequenceModifier: ...
    @property
    def rotation(self) -> SequenceModifier: ...
    @property
    def offset(self) -> TupleSequenceModifier: ...
    @property
    def scale(self) -> TupleSequenceModifier: ...
    @property
    def flip(self) -> SequenceModifier: ...

class TupleSequenceModifier(SequenceModifier):
    def set(self, val: tuple[int, int], filter: Callable[[Frame], bool] | None = None) -> TupleSequenceModifier: ...
    def add(self, val: tuple[int, int], filter: Callable[[Frame], bool] | None = None) -> TupleSequenceModifier: ...
    def mul(self, val: tuple[int, int], filter: Callable[[Frame], bool] | None = None) -> TupleSequenceModifier: ...
    def transform(self, transformer: Callable[[Frame, tuple[int, int]], tuple[int, int]]) -> TupleSequenceModifier: ...

    def __getitem__(self, key: int | slice) -> TupleSequenceModifier: ...

class FrameSequenceModifier(SequenceModifier):
    def reverse(self) -> FrameSequenceModifier: ...
    def transform(self, transformer: Callable[[Frame], Frame]) -> FrameSequenceModifier: ...
    def clsn1(self, clsn: Clsn, filter: Callable[[Frame], bool] | None = None) -> FrameSequenceModifier: ...
    def clsn2(self, clsn: Clsn, filter: Callable[[Frame], bool] | None = None) -> FrameSequenceModifier: ...
    def default(self, filter: Callable[[Frame], bool] | None = None) -> FrameSequenceModifier: ...

    def __getitem__(self, key: int | slice) -> FrameSequenceModifier: ...

class Sequence:
    _frames: list[Frame]

    def __init__(self, frames: list[Frame]): ...
    def compile(self) -> str: ...
    def extend(self, frames: Frame | list[Frame] | Sequence | SequenceModifier | FrameSequenceModifier | TupleSequenceModifier) -> Sequence: ...
    def __getitem__(self, key: int | slice) -> Sequence: ...

    @property
    def frames(self) -> FrameSequenceModifier: ...
    @property
    def group(self) -> SequenceModifier: ...
    @property
    def index(self) -> SequenceModifier: ...
    @property
    def length(self) -> SequenceModifier: ...
    @property
    def rotation(self) -> SequenceModifier: ...
    @property
    def offset(self) -> TupleSequenceModifier: ...
    @property
    def scale(self) -> TupleSequenceModifier: ...
    @property
    def flip(self) -> SequenceModifier: ...

class Animation:
    _frames: Sequence
    _id: int | None = None

    def __init__(self, frames: Sequence | SequenceModifier | list[Frame] | Frame | None = None, id: int | None = None, external: bool = False): ...
    def compile(self) -> str: ...

    @property
    def sequence(self) -> FrameSequenceModifier: ...

__all__ = ["Animation", "Sequence", "Frame", "AnimationFlip", "AnimationTrans"]


from typing import Optional

from mdk.types.context import CompilerContext
from mdk.types.expressions import Expression
from mdk.types.variables import VariableExpression
from mdk.types.builtins import BoolType, IntType, ShortType, ByteType
from mdk.types.defined import StateNoType, EnumType, FlagType
from mdk.types.specifier import TypeCategory

from mdk.stdlib.controllers import DisplayToClipboard, AppendToClipboard

from mdk.utils.shared import format_bool, convert
from mdk.utils.expressions import check_types_assignable

def TriggerAnd(*exprs_: Expression | bool):
    ## auto-convert bools to BoolExpression.
    exprs = list(exprs_)
    for index in range(len(exprs)):
        expr = exprs[index]
        if type(expr) == bool: exprs[index] = format_bool(expr)

    ## check all exprs as input have bool type
    ## for compatibility we must treat Int as Bool.
    for expr in exprs:
        if not isinstance(expr, Expression):
            raise Exception(f"Expected input to AND statement to be convertible to an Expression, but found {type(expr)}.")
        if expr.type not in [BoolType, IntType, ShortType, ByteType, StateNoType]:
            if (not isinstance(expr.type, EnumType) and not isinstance(expr.type, FlagType)) or not expr.type.register:
                raise Exception(f"Expected input to AND statement to be an Expression with type `bool` or an equivalent type, not {expr.type.name}.")
        
    expr_string = " && ".join([expr.exprn for expr in exprs if isinstance(expr, Expression)])
    return Expression(f"({expr_string})", BoolType)

def TriggerOr(*exprs_: Expression | bool):
    ## auto-convert bools to BoolExpression.
    exprs = list(exprs_)
    for index in range(len(exprs)):
        expr = exprs[index]
        if type(expr) == bool: exprs[index] = format_bool(expr)

    ## check all exprs as input have bool type
    ## for compatibility we must treat Int as Bool.
    for expr in exprs:
        if not isinstance(expr, Expression):
            raise Exception(f"Expected input to OR statement to be convertible to an Expression, but found {type(expr)}.")
        if expr.type not in [BoolType, IntType, ShortType, ByteType, StateNoType]:
            if (not isinstance(expr.type, EnumType) and not isinstance(expr.type, FlagType)) or not expr.type.register:
                raise Exception(f"Expected input to OR statement to be an Expression with type `bool` or an equivalent type, not {expr.type.name}.")
        
    expr_string = " || ".join([expr.exprn for expr in exprs if isinstance(expr, Expression)])
    return Expression(f"({expr_string})", BoolType)

def TriggerNot(expr: Expression):
    ## auto-convert bools to BoolExpression.
    if type(expr) == bool:
        expr = format_bool(expr)

    """
    Note: this is intentionally disabled. In CNS, it is fine to do something like `!Life` to check when life is 0, but Life as an IntExpression
    would not work with this check. I think it's useful and natural to write `if not Life` in Python, so we can retain this CNS behaviour.

    if not isinstance(expr, BoolExpression):
        raise Exception(f"First parameter to NOT statement should be BoolExpression or BoolVar, not {type(expr)}.")
    """

    return Expression(f"!({expr.exprn})", BoolType)

def TriggerNotIn(expr1: Expression, expr2: Expression):
    ## unfortunately, just using `expr1 in expr2` gives us a bool in return (i guess the language automatically calls __bool__ for us, which is annoying)
    ## so there's duplicated logic here from the __contains__ dunder.
    if not isinstance(expr1, Expression):
        expr1 = convert(expr1)
    if not isinstance(expr2, Expression):
        expr2 = convert(expr2)
    if expr1.type.category != TypeCategory.FLAG or expr2.type.category != TypeCategory.FLAG:
        raise Exception(f"Types {expr1.type.name} and {expr2.type.name} are not Flag types and cannot be tested for membership.")
    return Expression(f"({expr2.__str__()} != {expr1.__str__()})", BoolType)

def TriggerAssign(expr1: Expression, expr2: Expression):
    if not isinstance(expr1, VariableExpression):
        raise Exception(f"First parameter to assignment statement should be a VariableExpression, not {type(expr1)}.")
    if not isinstance(expr2, Expression):
        expr2 = convert(expr2)
    if check_types_assignable(expr1.type, expr2.type) == None:
        raise Exception(f"Inputs to assignment expression must have assignable types, not {expr1.type.name} and {expr2.type.name}.")

    return expr1.make_expression(f"({expr1.exprn} := {expr2.exprn})")

def TriggerPush(depth: Optional[int]):
    ctx = CompilerContext.instance()
    if not isinstance(ctx.current_trigger, Expression):
        ## it's legal for current_trigger to be None, it means the controller was called outside of an `if`.
        ## (i don't think it's possible for TriggerPush to invoke in this case, but better for safety).
        ctx.current_trigger = format_bool(True)
    
    if depth == None:
        ctx.trigger_stack.append(ctx.current_trigger)
    else:
        check_val = 0 if len(ctx.if_stack) == 0 else ctx.if_stack[-1]
        ## if check_val is 0, and depth is 1, we can reset mdk_internalTrigger
        ## as we know this is the opening of a new block.
        if check_val == 0 and depth == 1:
            ctx.check_stack.append(Expression(f"(mdk_internalTrigger := 0) || true", BoolType))
        else:
            ctx.check_stack.append(Expression(f"mdk_internalTrigger = {check_val}", BoolType))

        if depth != -1:
            ctx.check_stack.append(ctx.current_trigger)
            ctx.check_stack.append(Expression(f"(mdk_internalTrigger := {depth}) || true", BoolType))
            ctx.if_stack.append(depth)
        else:
            ctx.trigger_stack.append(Expression(f"mdk_internalTrigger = {check_val}", BoolType))
            ctx.if_stack = []
            ctx.check_stack = []

def TriggerPop(depth: Optional[int]):
    ctx = CompilerContext.instance()
    if depth == None:
        if len(ctx.trigger_stack) == 0:
            raise Exception(f"Tried to pop triggers from an empty trigger stack.")
        ctx.trigger_stack.pop()
    else:
        if len(ctx.if_stack) == 0: raise Exception(f"Tried to pop triggers from an empty trigger stack.")
        if len(ctx.if_stack) == 1: ctx.if_stack.pop()
        if len(ctx.if_stack) > 1:
            target = ctx.current_state
            if target == None: target = ctx.current_template
            if target == None: raise Exception(f"Trigger conditional applied outside of context.")

            if len(target.controllers) == 0: raise Exception(f"Trigger conditional applied with no state controllers.")

            target.controllers[-1].triggers.append(Expression(f"mdk_internalTrigger := {ctx.if_stack[-2]}", BoolType))
            ctx.if_stack.pop()

## this function is used if a bare `print` statement is found inside a `statedef`, `statefunc`, or `template`.
## it replaces the `print` with an appropriate `DisplayToClipboard`.
## this accepts an optional `append` kwarg, if this is set to `True` it will do AppendToClipboard instead.
def TriggerPrint(*args, **kwargs):
    ## before anything: if compile = False, just print normally
    if 'compile' in kwargs and kwargs['compile'] == False:
        print(*args)
        return

    ctx = CompilerContext.instance()
    if len(args) == 0: raise Exception("Must provide text input to convertible `print` statement.")

    if len(ctx.format_params) > 6:
        raise Exception("Convertible print statements and Clipboard controllers only support up to 6 parameters.")

    text: str = args[0]

    if 'end' in kwargs and kwargs['end'] != None:
        text += kwargs['end']

    ## handle formatting:
    ### - any % in the string becomes %%, UNLESS it matches a known type specifier.
    ### - any \ in the string becomes \\
    text = text.replace("\\", "\\\\").replace("%", "%%")
    text = text.replace("%%d", "%d").replace("%%f", "%f").replace("%%e", "%e").replace("%%E", "%E").replace("%%g", "%g").replace("%%G", "%G")

    if 'append' in kwargs and kwargs['append'] == True:
        AppendToClipboard(text = text, params = tuple(ctx.format_params) if len(ctx.format_params) > 0 else None)
    else:
        DisplayToClipboard(text = text, params = tuple(ctx.format_params) if len(ctx.format_params) > 0 else None)

    ctx.format_params = []


from typing import Optional, Callable, Union
from functools import partial
import inspect
import copy
import traceback
import os

import mtl.project
from mtl.types.translation import ForwardParameter, StateDefinitionScope, StateScopeType as MtlScopeType
import mtlcc

from mdk.types.context import StateDefinition, TemplateDefinition, StateController, CompilerContext, StateScope, TriggerDefinition, StateScopeType, TranslationMode
from mdk.types.specifier import TypeSpecifier
from mdk.types.errors import CompilationException
from mdk.types.expressions import Expression
from mdk.types.builtins import IntType, StateNoType, BoolType
from mdk.types.defined import StateType, StateTypeT, MoveType, MoveTypeT, PhysicsType, PhysicsTypeT, FloatPairType, EnumType, FlagType

from mdk.utils.shared import convert, convert_tuple, format_bool, create_compiler_error
from mdk.utils.compiler import write_controller, rewrite_function
from mdk.utils.animation import read_animations

from mdk.stdlib.controllers import ChangeState
from mdk.resources.animation import Animation

def build(
        def_file: str, output: str, 
        run_mtl: bool = True, skip_templates: bool = False, locations: bool = True, compress: bool = False, 
        preserve_ir: bool = False, target_folder: str = "mdk-out", debug_build: bool = False) -> None:
    context = CompilerContext.instance()
    context.debug_build = debug_build
    try:
        output_path = os.path.join(os.path.abspath(os.path.dirname(def_file)), output)
        print(f"Will build state definitions to output path {output_path}.")
        # load the DEF file early for anim loading.
        project = mtl.project.loadDefinition(def_file)
        ## load animations
        if project.anim_file != "":
            anim_path = os.path.join(os.path.abspath(os.path.dirname(def_file)), project.anim_file)
            print(f"Loading external animations from source animation file {anim_path}")
            old_animations = context.animations
            new_animations = read_animations(anim_path)
            ## we need to connect the external-marked anims in `old_animations` to their definitions in `new_animations`,
            ## and remove those from `new_animations`.
            for anim in old_animations:
                matches = [x for x in new_animations if x._id == anim._id]
                if anim._external and len(matches) == 1:
                    anim._frames = copy.deepcopy(matches[0]._frames)
                    anim._external = False
                    new_animations = [x for x in new_animations if x._id != anim._id]
                elif anim._external and len(matches) != 1:
                    raise Exception(f"Error while identifying externally-linked animation with ID {anim._id}; found {len(matches)} matching animations in AIR file")
            context.animations = old_animations + new_animations
        project.anim_file = output_path + ".air"
        with open(output_path + ".air", mode="w") as f:
            for anim in context.animations:
                if not anim._external:
                    f.write(anim.compile())
                    f.write("\n")

        ## builds the character from the input data.
        for state in context.statedefs:
            definition = context.statedefs[state]
            ## if an Animation object is forward-declared, read its assigned ID here
            ## and put it in the statedef's params.
            if definition._fwd_animation != None:
                if "anim" in definition.params and definition.params["anim"] != None:
                    raise Exception(f"Attempted to use an Animation object for state {definition} but an animation ID is already assigned!")
                definition.params["anim"] = Expression(str(definition._fwd_animation._id), IntType)
            context.current_state = definition
            try:
                definition.fn() # this call registers the controllers in the statedef with the statedef itself.
                if context.debug_build:
                    virtual = StateController()
                    virtual.location = definition.location
                    virtual.type = "Null"
                    virtual.triggers = [Expression("true", BoolType)]
                    definition.controllers.insert(0, virtual)
            except Exception as exc:
                raise CompilationException(exc)
            context.current_state = None

        lib_groups = set()
        lib_targets: list[Callable | TypeSpecifier] = []
        if not skip_templates and len(context.templates) != 0:
            for t in context.templates:
                lib_targets.append(context.templates[t].fn)
                if context.templates[t].library == None:
                    context.templates[t].library = output + ".inc"
                lib_groups.add(context.templates[t].library)
        if not skip_templates and len(context.triggers) != 0:
            for t in context.triggers:
                lib_targets.append(context.triggers[t].fn)
                if context.triggers[t].library == None:
                    context.triggers[t].library = output + ".inc"
                lib_groups.add(context.triggers[t].library)
        if not skip_templates and len(context.typedefs) != 0:
            for t in context.typedefs:
                if not context.typedefs[t].register: continue
                lib_targets.append(context.typedefs[t])
                if context.typedefs[t].library == None:
                    context.typedefs[t].library = output + ".inc"
                lib_groups.add(context.typedefs[t].library)

        if len(lib_targets) > 0: 
            library(lib_targets, dirname = os.path.abspath(os.path.dirname(def_file)), locations = locations)
        
        with open(output, mode="w") as f:
            if not skip_templates and len(lib_groups) != 0:
                for group in lib_groups:
                    f.write("[Include]\n")
                    f.write(f"source = {group}\n")
            if not compress: f.write("\n")
            for name in context.statedefs:
                statedef = context.statedefs[name]
                f.write(f"[Statedef {name}]\n")
                for param in statedef.params:
                    f.write(f"{param} = {statedef.params[param].__str__()}\n")
                for local in statedef.locals:
                    f.write(f"local = {local.name} = {local.type.name}\n")
                if statedef.scope != None:
                    f.write(f"scope = {statedef.scope}\n")
                file_path = inspect.getsourcefile(statedef.fn)
                _, line_number = inspect.getsourcelines(statedef.fn)
                if locations:
                    f.write(f"mtl.location.file = {file_path}\n")
                    f.write(f"mtl.location.line = {line_number}\n")
                if not compress: f.write("\n")
                for controller in statedef.controllers:
                    write_controller(controller, f, locations)
                    if not compress: f.write("\n")
                if not compress: f.write("\n")

        ## read the DEF file (using MTL) and add the output to the statefile list,
        ## then re-export the DEF file.
        if run_mtl:
            print(f"Preparing to run MTL compiler for input project {def_file}.")
            project.source_files.append(output)
            for global_variable in context.globals:
                scoped = StateDefinitionScope(MtlScopeType.SHARED, None)
                if global_variable.scope != None:
                    if global_variable.scope.scope == StateScopeType.PLAYER:
                        scoped = StateDefinitionScope(MtlScopeType.PLAYER, None)
                    if global_variable.scope.scope == StateScopeType.TARGET:
                        scoped = StateDefinitionScope(MtlScopeType.TARGET, None)
                    if global_variable.scope.scope == StateScopeType.HELPER:
                        scoped = StateDefinitionScope(MtlScopeType.HELPER, global_variable.scope.target)
                project.global_forwards.append(ForwardParameter(global_variable.name, global_variable.type.name, is_system = global_variable.is_system, scope = scoped))
            ## system globals
            project.global_forwards.append(ForwardParameter("mdk_internalTrigger", "int", is_system = False, scope = StateDefinitionScope(MtlScopeType.SHARED, None)))
            mtlcc.runCompilerFromDef(def_file, os.path.join(os.path.abspath(os.path.dirname(def_file)), target_folder), project)

        ## delete the output file if we're not preserving IR
        if not preserve_ir: 
            os.remove(output)
            for name in lib_groups:
                os.remove(name)
    except CompilationException as exc:
        create_compiler_error(exc)
    except Exception as exc:
        print("An internal error occurred while compiling a template, bug the developers.")
        raise exc

def library(inputs: list[Callable[..., None] | TypeSpecifier], dirname: str = "", output: Optional[str] = None, locations: bool = True, preserve_ir: bool = False) -> None:
    if len(inputs) == 0:
        raise Exception("Please specify some templates and/or triggers to be built.")
    context = CompilerContext.instance()
    try:
        per_file: dict[str, list[Union[TriggerDefinition, TemplateDefinition, TypeSpecifier]]] = {}
        for template in context.templates:
            definition = context.templates[template]
            context.current_template = definition
            kwargs = {}
            for param in definition.params:
                t = definition.params[param]
                kwargs[param] = Expression(param, t)
            try:
                definition.fn(**kwargs)
            except Exception as exc:
                raise CompilationException(exc)
            context.current_template = None
            if definition.library == None and output == None:
                raise CompilationException(f"Output library was not specified for template {definition.fn.__name__}, either specify a default output in call to library() or specify a library in the template() annotation.")
            if definition.library == None and output != None and definition.fn in inputs:
                definition.library = output
            if definition.library not in per_file:
                per_file[definition.library] = [] # type: ignore
            per_file[definition.library].append(definition) # type: ignore

        for trigger in context.triggers:
            definition = context.triggers[trigger]
            kwargs = {}
            for param in definition.params:
                t = definition.params[param]
                kwargs[param] = Expression(param, t)
            try:
                definition.exprn = definition.fn(**kwargs)
            except Exception as exc:
                raise CompilationException(exc)
            if definition.library == None and output != None and definition.fn in inputs:
                definition.library = output
            if definition.library not in per_file:
                per_file[definition.library] = [] # type: ignore
            per_file[definition.library].append(definition) # type: ignore

        for typedef in context.typedefs:
            definition = context.typedefs[typedef]
            if not definition.register:
                continue
            if not isinstance(definition, EnumType) and not isinstance(definition, FlagType):
                raise CompilationException(f"Can only emit type declarations for EnumType and FlagType, not {type(definition)}.")
            if definition.library == None and output != None and definition.fn in inputs:
                definition.library = output
            if definition.library not in per_file:
                per_file[definition.library] = [] # type: ignore
            per_file[definition.library].append(definition) # type: ignore

        for group in per_file:
            group_path = os.path.join(dirname, group)
            print(f"Creating output include file at path {group_path}.")
            with open(group_path, mode="w") as f:
                for definition in per_file[group]:
                    if isinstance(definition, TemplateDefinition):
                        f.write("[Define Template]\n")
                        f.write(f"name = {definition.fn.__name__}\n")
                        for local in definition.locals:
                            f.write(f"local = {local.name} = {local.type.name}\n")
                        file_path = inspect.getsourcefile(definition.fn)
                        _, line_number = inspect.getsourcelines(definition.fn)
                        if locations:
                            f.write(f"mtl.location.file = {file_path}\n")
                            f.write(f"mtl.location.line = {line_number}\n")
                        f.write("\n")
                        f.write("[Define Parameters]\n")
                        for param in definition.params:
                            f.write(f"{param} = {definition.params[param].name}\n")
                        f.write("\n")
                        for controller in definition.controllers:
                            write_controller(controller, f, locations)
                            f.write("\n")
                        f.write("\n")
                    elif isinstance(definition, TriggerDefinition):
                        f.write("[Define Trigger]\n")
                        f.write(f"name = {definition.fn.__name__}\n")
                        f.write(f"type = {definition.result.name}\n")
                        if not isinstance(definition.exprn, Expression):
                            raise Exception(f"Trigger definition {definition.fn.__name__} returned {type(definition.exprn)}, but must return Expression instead.")
                        f.write(f"value = {definition.exprn.exprn}\n")
                        file_path = inspect.getsourcefile(definition.fn)
                        _, line_number = inspect.getsourcelines(definition.fn)
                        if locations:
                            f.write(f"mtl.location.file = {file_path}\n")
                            f.write(f"mtl.location.line = {line_number}\n")
                        f.write("\n")
                        f.write("[Define Parameters]\n")
                        for param in definition.params:
                            f.write(f"{param} = {definition.params[param].name}\n")
                        f.write("\n")
                    elif isinstance(definition, TypeSpecifier):
                        f.write("[Define Type]\n")
                        f.write(f"name = {definition.name}\n")
                        if isinstance(definition, EnumType):
                            f.write(f"type = enum\n")
                            for member in definition.members:
                                f.write(f"enum = {member}\n")
                        elif isinstance(definition, FlagType):
                            f.write(f"type = flag\n")
                            for member in definition.members:
                                f.write(f"flag = {member}\n")
                        f.write("\n")
    except CompilationException as exc:
        create_compiler_error(exc)
    except Exception as exc:
        print("An internal error occurred while compiling a template, bug the developers.")
        raise exc
    
## very simple decorator to ensure a function called from a statedef
## can also have triggers applied correctly.
def statefunc(mode: TranslationMode = TranslationMode.STANDARD) -> Callable[[Callable[..., None]], Callable[..., None]]:
    def wrapped(fn: Callable[..., None]) -> Callable[..., None]:
        ctx = CompilerContext.instance()
        new_fn = rewrite_function(fn, mode = mode)
        ctx.statefuncs.append(fn.__name__)
        def inner(*args, **kwargs):
            if ctx.debug_build and ctx.current_state != None:
                virtual = StateController()
                callsite = traceback.extract_stack()[-2]
                virtual.location = (callsite.filename, callsite.lineno if callsite.lineno != None else 0)
                virtual.type = "Null"
                virtual.triggers = [Expression("true", BoolType)]
                ctx.current_state.controllers.append(virtual)
            new_fn(*args, **kwargs)
        return inner
    return wrapped

def statedef(
    type: Optional[StateType] = None,
    movetype: Optional[MoveType] = None,
    physics: Optional[PhysicsType] = None,
    anim: Optional[int | Animation] = None,
    velset: Optional[tuple[float, float]] = None,
    ctrl: Optional[bool] = None,
    poweradd: Optional[int] = None,
    juggle: Optional[int] = None,
    facep2: Optional[bool] = None,
    hitdefpersist: Optional[bool] = None,
    movehitpersist: Optional[bool] = None,
    hitcountpersist: Optional[bool] = None,
    sprpriority: Optional[int] = None,
    stateno: Optional[int] = None,
    scope: Optional[StateScope] = None,
    mode: TranslationMode = TranslationMode.STANDARD
) -> Callable[[Callable[[], None]], Callable[..., StateController]]:
    def decorator(fn: Callable[[], None]) -> Callable[..., StateController]:
        return create_statedef(fn, type, movetype, physics, anim, velset, ctrl, poweradd, juggle, facep2, hitdefpersist, movehitpersist, hitcountpersist, sprpriority, stateno, scope, mode)
    return decorator

## used by the @statedef decorator to create new statedefs,
## but can also be used by character developers to create statedefs ad-hoc (e.g. in a loop).
def create_statedef(
    fn: Callable[[], None],
    type: Optional[StateType] = None,
    movetype: Optional[MoveType] = None,
    physics: Optional[PhysicsType] = None,
    anim: Optional[int | Animation] = None,
    velset: Optional[tuple[float, float]] = None,
    ctrl: Optional[bool] = None,
    poweradd: Optional[int] = None,
    juggle: Optional[int] = None,
    facep2: Optional[bool] = None,
    hitdefpersist: Optional[bool] = None,
    movehitpersist: Optional[bool] = None,
    hitcountpersist: Optional[bool] = None,
    sprpriority: Optional[int] = None,
    stateno: Optional[int] = None,
    scope: Optional[StateScope] = None,
    mode: TranslationMode = TranslationMode.STANDARD
) -> Callable[..., StateController]:
    print(f"Discovered a new StateDef named {fn.__name__}. Will process and load this StateDef.")
    
    new_fn = rewrite_function(fn, mode = mode)
    callsite = traceback.extract_stack()[-3]
    statedef = StateDefinition(new_fn, {}, [], [], (callsite.filename, callsite.lineno + 1 if callsite.lineno != None else 0), scope, None)

    # apply each parameter
    if stateno != None: statedef.params["id"] = Expression(str(stateno), IntType)
    if type != None:
        statedef.params["type"] = Expression(type.name, StateTypeT)
    if movetype != None:
        statedef.params["movetype"] = Expression(movetype.name, MoveTypeT)
    if physics != None:
        statedef.params["physics"] = Expression(physics.name, PhysicsTypeT)
    if anim != None: 
        ## anim can be an Animation, in which case the `anim` param should be set to the Animation's ID.
        ## however, this step of statedef loading occurs pre-compilation for animations, so the Animation object
        ## may not have an assigned ID yet.
        ## in this case, we store the Animation in the `_fwd_animation` property, and when the state is compiled,
        ## the ID gets read and set.
        if isinstance(anim, Animation):
            statedef._fwd_animation = anim
        else:
            statedef.params["anim"] = Expression(str(anim), IntType)
    if velset != None: statedef.params["velset"] = convert_tuple(velset, FloatPairType)
    if ctrl != None: statedef.params["ctrl"] = format_bool(ctrl)
    if poweradd != None: statedef.params["poweradd"] = Expression(str(poweradd), IntType)
    if juggle != None: statedef.params["juggle"] = Expression(str(juggle), IntType)
    if facep2 != None: statedef.params["facep2"] = format_bool(facep2)
    if hitdefpersist != None: statedef.params["hitdefpersist"] = format_bool(hitdefpersist)
    if movehitpersist != None: statedef.params["movehitpersist"] = format_bool(movehitpersist)
    if hitcountpersist != None: statedef.params["hitcountpersist"] = format_bool(hitcountpersist)
    if sprpriority != None: statedef.params["sprpriority"] = Expression(str(sprpriority), IntType)

    # add the new statedef to the context
    ctx = CompilerContext.instance()
    if fn.__name__ in ctx.statedefs:
        raise Exception(f"Attempted to overwrite statedef with name {fn.__name__}.")
    ctx.statedefs[fn.__name__] = statedef

    return partial(ChangeState, value = fn.__name__)

def do_template(name: str, validator: Optional[Callable[..., dict[str, Expression]]], *args, **kwargs) -> StateController:
    def generic_template(*args, **kwargs) -> StateController:
        if len(args) != 0:
            raise Exception("Templates cannot be called with positional arguments, only keyword arguments.")
        ## if a validator was provided, we give the template author a chance to validate and even modify input arguments.
        if validator != None and (kwargs := validator(**kwargs)) == None:
            raise Exception(f"Could not place call to template with name {name}, template validation check failed.")
        context = CompilerContext.instance()

        ## need to check for any StateNoType inputs
        target_template = context.templates[name]

        new_controller = StateController()
        new_controller.type = name
        new_controller.params = {}
        for arg in kwargs:
            target_type = target_template.params[arg]
            next_arg = kwargs[arg]
            if target_type == StateNoType and isinstance(next_arg, partial):
                new_controller.params[arg] = Expression(next_arg.keywords["value"], StateNoType)
            elif target_type == StateNoType and isinstance(next_arg, Callable):
                new_controller.params[arg] = Expression(next_arg.__name__, StateNoType)
            else:
                new_controller.params[arg] = kwargs[arg]
        if len(context.trigger_stack) != 0 or len(context.if_stack) != 0:
            new_controller.triggers = copy.deepcopy(context.trigger_stack)
            if len(context.if_stack) != 0:
                new_controller.triggers.append(Expression(f"mdk_internalTrigger = {context.if_stack[-1]}", BoolType))
            if len(context.check_stack) != 0:
                new_controller.triggers = copy.deepcopy(context.check_stack) + new_controller.triggers
                context.check_stack = []
        else:
            new_controller.triggers = [format_bool(True)]
        ## i believe this should always be right, since it flows from <callsite> -> do_template -> generic_template
        callsite = traceback.extract_stack()[-3]
        new_controller.location = (callsite.filename, callsite.lineno if callsite.lineno != None else 0)
        return new_controller
    ctrl = generic_template(*args, **kwargs)
    ctrl.type = name
    context = CompilerContext.instance()
    if context.current_state != None:
        context.current_state.controllers.append(ctrl)
    elif context.current_template != None:
        context.current_template.controllers.append(ctrl)
    return ctrl

def do_trigger(name: str, validator: Optional[Callable[..., dict[str, Expression]]], *args, **kwargs) -> Expression:
    if len(kwargs) != 0:
        raise Exception("Triggers cannot be called with keyword arguments, only positional arguments.")
    ## if a validator was provided, we give the trigger author a chance to validate and even modify input arguments.
    if validator != None and (args := validator(*args)) == None:
        raise Exception(f"Could not place call to trigger with name {name}, trigger validation check failed.")
    context = CompilerContext.instance()
    param_string = ", ".join([convert(arg).exprn for arg in args])
    ## just trust the output type. the MTL side will do final type validation during trigger replacement.
    return Expression(f"{name}({param_string})", context.triggers[name].result)

def template(inputs: list[TypeSpecifier] = [], library: Optional[str] = None, validator: Optional[Callable[..., dict[str, Expression]]] = None) -> Callable[[Callable[..., None]], Callable[..., StateController]]:
    def decorator(fn: Callable[..., None]) -> Callable[..., StateController]:
        print(f"Discovered a new Template named {fn.__name__}. Will process and load this Template.")
        # get params of decorated function
        signature = inspect.signature(fn)

        # create new function with ast fixes
        new_fn = rewrite_function(fn)

        # ensure parameters align
        if len(signature.parameters) != len(inputs):
            raise Exception(f"Mismatch in template parameter count: saw {len(inputs)} input types, and {len(signature.parameters)} real parameters.")

        params: dict[str, TypeSpecifier] = {}
        index = 0
        for param in signature.parameters:
            params[param] = inputs[index]
            index += 1
        template = TemplateDefinition(new_fn, library, params, [], [])

        # add the new template to the context
        ctx = CompilerContext.instance()
        if fn.__name__ in ctx.templates:
            raise Exception(f"Attempted to overwrite template with name {fn.__name__}.")
        ctx.templates[fn.__name__] = template

        return partial(do_template, fn.__name__, validator)
    return decorator

def trigger(inputs: list[TypeSpecifier], result: TypeSpecifier, library: Optional[str] = None, validator: Optional[Callable[..., dict[str, Expression]]] = None) -> Callable[[Callable[..., Expression]], Callable[..., Expression]]:
    def decorator(fn: Callable[..., Expression]) -> Callable[..., Expression]:
        print(f"Discovered a new Trigger named {fn.__name__}. Will process and load this Trigger.")
        # get params of decorated function
        signature = inspect.signature(fn)

        # create new function with ast fixes
        ## note: triggers cannot overload print functions (because triggers can't contain controllers).
        new_fn = rewrite_function(fn, overload_print = False) # type: ignore

        # ensure parameters align
        if len(signature.parameters) != len(inputs):
            raise Exception(f"Mismatch in trigger parameter count: saw {len(inputs)} input types, and {len(signature.parameters)} real parameters.")

        params: dict[str, TypeSpecifier] = {}
        index = 0
        for param in signature.parameters:
            params[param] = inputs[index]
            index += 1

        trigger = TriggerDefinition(new_fn, library, result, params) # type: ignore

        # add the new template to the context
        ctx = CompilerContext.instance()
        if fn.__name__ in ctx.triggers:
            raise Exception(f"Attempted to overwrite trigger with name {fn.__name__}.")
        ctx.triggers[fn.__name__] = trigger

        return partial(do_trigger, fn.__name__, validator)
    return decorator

class ControllerProps:
    def __init__(self, ignorehitpause: Expression | bool | None = None, persistent: Expression | int | None = None):
        self.ignorehitpause = convert(ignorehitpause) if ignorehitpause != None else None
        self.persistent = convert(persistent) if persistent != None else None
        self.prev_state = (None, None)
    def __enter__(self):
        context = CompilerContext.instance()
        self.prev_state = context.default_state
        context.default_state = (self.ignorehitpause, self.persistent)
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        context = CompilerContext.instance()
        context.default_state = self.prev_state
        return False
    
__all__ = ["build", "library", "statedef", "create_statedef", "template", "trigger"]


from typing import Optional

from mdk.types.specifier import TypeSpecifier
from mdk.types.builtins import StringType, UStringType

## this is based on the MTL type conversion function in mtl.utils.compiler.
def check_types_assignable(spec1: TypeSpecifier, spec2: TypeSpecifier) -> Optional[TypeSpecifier]:
    ## if types match, t1 return the type.
    if spec1 == spec2: return spec1

    ## 'any' type can be converted to anything.
    if spec1.name == "any": return spec2
    if spec2.name == "any": return spec1

    ## ustring and string are convertible to ustring.
    if spec1 in [StringType, UStringType] and spec2 in [StringType, UStringType]:
        return UStringType
    
    ## `int` is implicitly convertible to `anim`, `stateno`
    if spec1.name == "int" and spec2.name in ["anim", "stateno"]:
        return spec2
    if spec1.name in ["stateno", "anim"] and spec2.name == "int":
        return spec1

    ## `int` is implicitly convertible to `float`
    if spec1.name == "int" and spec2.name == "float":
        return spec2
    
    ## allow this, because Python is pretty permissive with numerics.
    if spec1.name == "float" and spec2.name == "int":
        return spec1
        #raise Exception("Conversion from float to int may result in loss of precision. If this is intended, use functions like ceil or floor to convert, or explicitly cast one side of the expression.")
    
    ## widening
    if spec1.name == "bool" and spec2.name in ["byte", "short", "int"]:
        return spec2
    if spec1.name == "byte" and spec2.name in ["short", "int"]:
        return spec2
    if spec1.name == "short" and spec2.name in ["int"]:
        return spec2
    
    ## implicit char -> byte
    if spec1.name == "char" and spec2.name in ["byte"]:
        return spec2

    return None

def check_any_assignable(spec1: TypeSpecifier, spec2: list[TypeSpecifier]) -> Optional[TypeSpecifier]:
    for s in spec2:
        if (assgn := check_types_assignable(spec1, s)) != None:
            return assgn
    return None


from mdk.types.builtins import BoolType as BoolType
from mdk.types.context import CompilerContext as CompilerContext
from mdk.types.expressions import Expression as Expression
from mdk.types.variables import VariableExpression as VariableExpression
from mdk.utils.expressions import check_types_assignable as check_types_assignable
from mdk.utils.shared import format_bool as format_bool

def TriggerAnd(expr1: Expression, expr2: Expression): ...
def TriggerOr(expr1: Expression, expr2: Expression): ...
def TriggerNot(expr: Expression): ...
def TriggerAssign(expr1: Expression, expr2: Expression): ...
def TriggerPush(): ...
def TriggerPop(): ...



from enum import Enum, Flag, auto, EnumType
from typing import TypeVar

__all__ = [
    "StateType", "MoveType", "PhysicsType", "HitType", "HitAttr", "TransType", "AssertType",
    "WaveType", "HelperType", "TeamType", "HitAnimType", "AttackType", "PriorityType", "PosType",
    "SpaceType", "TeamModeType", "HitFlagType", "GuardFlagType",
    "CompositeFlag"
]

## this TypeVar is a cute falsehood for the type checker.
## in implementation this takes no generic and __getattr__ returns a CompositeFlag.
## however, this results in usage of the flag having type complaints (because e.g. HitType.SCA returns a CompositeFlag which is wrong).
## therefore the `.pyi` uses a generic for the return type to 'trick' the type checker.
T = TypeVar('T', bound=Flag)
class CompositeFlag[T](EnumType):
    def __getattr__(self, name: str) -> T: ...

class StateType(Enum):
    """State type defined in statedef properties."""
    S = 0
    C = 1
    A = 2
    L = 3
    U = 4

class MoveType(Enum):
    """Move type (attack/idle/hurt) defined in statedef properties."""
    A = 0
    I = 1
    H = 2
    U = 3

class PhysicsType(Enum):
    """Physics type defined in statedef properties."""
    S = 0
    C = 1
    A = 2
    N = 3
    U = 4

class HitType(Flag, metaclass=CompositeFlag['HitType']):
    """Hit type as defined in e.g. a HitDef attr block."""
    S = 1
    C = auto()
    A = auto()

class HitAttr(Flag, metaclass=CompositeFlag['HitAttr']):
    """Hit type as defined in e.g. a HitDef attr block."""
    N = 1
    S = auto()
    H = auto()
    A = auto()
    T = auto()
    P = auto()

class TransType(Enum):
    """Type of transparency to apply in a visual display."""
    Add = 0
    Add1 = 1
    AddAlpha = 2
    Sub = 3
    none = 4

class AssertType(Enum):
    """Assertions which can be applied via the AssertSpecial controller."""
    Intro = 0
    Invisible = 1
    RoundNotOver = 2
    NoBarDisplay = 3
    NoBG = 4
    NoFG = 5
    NoStandGuard = 6
    NoCrouchGuard = 7
    NoAirGuard = 8
    NoAutoTurn = 9
    NoJuggleCheck = 10
    NoKOSnd = 11
    NoKOSlow = 12
    NoKO = 13
    NoShadow = 14
    GlobalNoShadow = 15
    NoMusic = 16
    NoWalk = 17
    TimerFreeze = 18
    Unguardable = 19

class WaveType(Enum):
    """Wave type used for ForceFeedback controllers."""
    Sine = 0
    Square = 1
    SineSquare = 2
    Off = 3

class HelperType(Enum):
    """Type of helper to be spawned from a Helper controller."""
    Normal = 0
    Player = 1
    Proj = 2

class TeamType(Enum):
    """Target team type for controllers such as HitDef."""
    E = 0
    F = 1
    B = 2

class HitAnimType(Enum):
    """Animation type to use when the target is hit by a HitDef or similar."""
    Light = 0
    Medium = 1
    Hard = 2
    Back = 3
    Up = 4
    DiagUp = 5

class AttackType(Enum):
    """The type of the attack (indicates where the attack will land)."""
    High = 0
    Low = 1
    Trip = 2
    none = 3

class PriorityType(Enum):
    """Priority of an attack, for clashing HitDefs."""
    Hit = 0
    Miss = 1
    Dodge = 2

class PosType(Enum):
    """Indicates position to base offsets from in e.g. Explod, Helper controllers."""
    P1 = 0
    P2 = 1
    Front = 3
    Back = 4
    Left = 5
    Right = 6
    none = 7

class SpaceType(Enum):
    """(1.1 only) indicates the space to use when calculating positions for e.g. Explod, Helper controllers."""
    Screen = 0
    Stage = 1

class TeamModeType(Enum):
    """Indicates the team mode which is currently in use (for TeamMode trigger)."""
    Single = 0
    Simul = 1
    Turns = 2

class HitFlagType(Flag, metaclass=CompositeFlag['HitFlagType']):
    """Used for a hitflag in a HitDef controller."""
    H = 1
    L = auto()
    A = auto()
    F = auto()
    D = auto()
    M = auto()

class GuardFlagType(Flag, metaclass=CompositeFlag['GuardFlagType']):
    """Used for a guardflag in a HitDef controller."""
    H = 1
    L = auto()
    A = auto()
    M = auto()


from mdk.types.builtins import BoolType as BoolType, FloatType as FloatType, IntType as IntType, StateNoType as StateNoType, StringType as StringType
from mdk.types.context import CompilerContext as CompilerContext, StateController as StateController
from mdk.types.defined import TupleType as TupleType
from mdk.types.expressions import Expression as Expression, TupleExpression as TupleExpression
from mdk.types.specifier import TypeSpecifier as TypeSpecifier
from mdk.utils.expressions import check_any_assignable as check_any_assignable
from mdk.utils.shared import convert as convert, format_bool as format_bool, format_tuple as format_tuple
from typing import Callable

def controller(**typeinfo) -> Callable[[Callable[..., StateController]], Callable[..., StateController]]: ...
def make_controller(fn, *args, typeinfo: dict[str, list[TypeSpecifier | None]], ignorehitpause: bool | None = None, persistent: int | None = None, **kwargs) -> StateController: ...



from mdk.types.builtins import *
from mdk.types.defined import *
from mdk.types.expressions import *
from mdk.types.variables import *
from mdk.types.enums import *
from mdk.types.context import StateScope as StateScope, StateScopeType as StateScopeType, SCOPE_HELPER as SCOPE_HELPER, SCOPE_PLAYER as SCOPE_PLAYER, SCOPE_TARGET as SCOPE_TARGET
from mdk.types.specifier import TypeSpecifier as TypeSpecifier



from typing import Callable
from mdk.types.expressions import Expression
from mdk.types.specifier import TypeSpecifier
from mdk.types.builtins import IntType, BoolType, FloatType, UStringType, StringType, AnyType
from mdk.types.defined import HitStringType
import mdk.types.defined as defined

from mdk.utils.shared import convert
from mdk.utils.expressions import check_types_assignable

## helper function to take 1 argument and the types involved and produce an output.
def TriggerExpression(name: str, inputs: list[TypeSpecifier], output: TypeSpecifier) -> Callable[..., Expression]:
    def _callable(*args) -> Expression:
        if len(args) != len(inputs):
            raise Exception(f"Trigger expression {name} expected {len(inputs)} inputs, but got {len(args)} instead.")
        conv_args: list[Expression] = []
        for index in range(len(args)):
            next_arg = convert(args[index])
            if not isinstance(next_arg, Expression):
                raise Exception(f"Inputs to trigger expressions should always be Expressions, not {type(args[index])}.")
            if next_arg.type in [UStringType, StringType] and inputs[index] in [UStringType, StringType]:
                ## UString and String are considered matching.
                next_arg.type = UStringType
            elif next_arg.type == IntType and inputs[index] == FloatType:
                next_arg.type = FloatType
            elif next_arg.type != inputs[index]:
                raise Exception(f"Trigger expression {name} expected input at index {index + 1} to have type {inputs[index].name}, not {next_arg.type.name}")
            conv_args.append(next_arg)
        return Expression(f"{name}({', '.join([str(arg) for arg in conv_args])})", output)
    return _callable

## helper function to take an optional argument and the types involved and produce an output.
def TriggerExpressionWithOptional(name: str, inputs: list[TypeSpecifier], output: TypeSpecifier) -> Callable[..., Expression]:
    def _callable(*args) -> Expression:
        if len(args) == 0:
            return Expression(f"{name}", output)
        if len(args) != len(inputs):
            raise Exception(f"Trigger expression {name} expected {len(inputs)} or 0 inputs, but got {len(args)} instead.")
        conv_args: list[Expression] = []
        for index in range(len(args)):
            next_arg = convert(args[index])
            if not isinstance(next_arg, Expression):
                raise Exception(f"Inputs to trigger expressions should always be Expressions, not {type(args[index])}.")
            if next_arg.type in [UStringType, StringType] and inputs[index] in [UStringType, StringType]:
                ## UString and String are considered matching.
                next_arg.type = UStringType
            elif next_arg.type != inputs[index]:
                raise Exception(f"Trigger expression {name} expected input at index {index + 1} to have type {inputs[index].name}, not {next_arg.type.name}")
            conv_args.append(next_arg)
        return Expression(f"{name}({', '.join([str(arg) for arg in conv_args])})", output)
    return _callable

## helper function for cond/ifelse specifically.
def TriggerExpressionCond(name: str) -> Callable[..., Expression]:
    def _callable(*args) -> Expression:
        if len(args) != 3:
            raise Exception(f"Trigger expression {name} requires 3 inputs, but got {len(args)} instead.")
        conv_args: list[Expression] = []
        for index in range(len(args)):
            next_arg = convert(args[index])
            if not isinstance(next_arg, Expression):
                raise Exception(f"Inputs to trigger expressions should always be Expressions, not {type(args[index])}.")
            conv_args.append(next_arg)
        if conv_args[0].type != BoolType:
            raise Exception(f"First argument to Cond or ifelse must be a bool, not {conv_args[0].type.name}.")
        if (output := check_types_assignable(conv_args[1].type, conv_args[2].type)) == None:
            raise Exception(f"Second and third arguments to Cond or ifelse must have assignable types; got {conv_args[1].type.name} and {conv_args[2].type.name}.")
        return Expression(f"{name}({', '.join([str(arg) for arg in conv_args])})", output)
    return _callable

## this is used for built-in position/velocity structures (e.g. Vel, Pos)
## it avoids the need for custom structure types, which are still not
## fully implemented either here or in MTL.
class PositionExpression:
    x: Expression
    y: Expression
    _name: str
    def __init__(self, name: str, type: TypeSpecifier):
        self._name = name
        self.x = Expression(f"{name} X", type)
        self.y = Expression(f"{name} Y", type)

Abs = TriggerExpression("abs", [FloatType], FloatType)
Acos = TriggerExpression("acos", [FloatType], FloatType)
AiLevel = Expression("AiLevel", IntType)
Alive = Expression("Alive", BoolType)
Anim = Expression("Anim", IntType)
## AnimElem
AnimElemNo = TriggerExpression("AnimElemNo", [IntType], IntType)
AnimElemTime = TriggerExpression("AnimElemTime", [IntType], IntType)
AnimExist = TriggerExpression("AnimExist", [IntType], BoolType)
AnimTime = Expression("AnimTime", IntType)
Asin = TriggerExpression("asin", [FloatType], FloatType)
Atan = TriggerExpression("atan", [FloatType], FloatType)
AuthorName = Expression("AuthorName", UStringType)
BackEdgeBodyDist = Expression("BackEdgeBodyDist", FloatType)
BackEdgeDist = Expression("BackEdgeDist", FloatType)
CanRecover = Expression("CanRecover", BoolType)
Ceil = TriggerExpression("ceil", [FloatType], IntType)
Command = Expression("Command", UStringType)
Cond = TriggerExpressionCond("Cond")
Const = TriggerExpression("Const", [UStringType], FloatType)
Const240p = TriggerExpression("Const240p", [FloatType], FloatType)
Const480p = TriggerExpression("Const480p", [FloatType], FloatType)
Const720p = TriggerExpression("Const720p", [FloatType], FloatType)
Cos = TriggerExpression("cos", [FloatType], FloatType)
Ctrl = Expression("Ctrl", BoolType)
DrawGame = Expression("DrawGame", BoolType)
E = Expression("E", FloatType)
Exp = TriggerExpression("exp", [FloatType], FloatType)
Facing = Expression("Facing", IntType)
Floor = TriggerExpression("floor", [FloatType], IntType)
FrontEdgeBodyDist = Expression("FrontEdgeBodyDist", FloatType)
FrontEdgeDist = Expression("FrontEdgeDist", FloatType)
## fvar
GameHeight = Expression("GameHeight", FloatType)
GameTime = Expression("GameTime", IntType)
GameWidth = Expression("GameWidth", IntType)
GetHitVar = TriggerExpression("GetHitVar", [UStringType], FloatType)
HitCount = Expression("HitCount", IntType)
HitDefAttr = Expression("HitDefAttr", HitStringType)
HitFall = Expression("HitFall", BoolType)
HitOver = Expression("HitOver", BoolType)
HitPauseTime = Expression("HitPauseTime", IntType)
HitShakeOver = Expression("HitShakeOver", BoolType)
HitVel = PositionExpression("HitVel", FloatType)
ID = Expression("ID", IntType)
ifelse = TriggerExpressionCond("ifelse")
InGuardDist = Expression("InGuardDist", BoolType)
IsHelper = TriggerExpressionWithOptional("IsHelper", [IntType], BoolType)
IsHomeTeam = Expression("IsHomeTeam", BoolType)
Life = Expression("Life", IntType)
LifeMax = Expression("LifeMax", IntType)
Ln = TriggerExpression("ln", [FloatType], FloatType)
Log = TriggerExpression("log", [FloatType], FloatType)
Lose = Expression("Lose", BoolType)
MatchNo = Expression("MatchNo", IntType)
MatchOver = Expression("MatchOver", BoolType)
MoveContact = Expression("MoveContact", IntType)
MoveGuarded = Expression("MoveGuarded", IntType)
MoveHit = Expression("MoveHit", IntType)
MoveType = Expression("MoveType", defined.MoveTypeT)
MoveReversed = Expression("MoveReversed", IntType)
Name = Expression("Name", UStringType)
NumEnemy = Expression("NumEnemy", IntType)
NumExplod = TriggerExpressionWithOptional("NumExplod", [IntType], IntType)
NumHelper = TriggerExpressionWithOptional("NumHelper", [IntType], IntType)
NumPartner = Expression("NumPartner", IntType)
NumProj = Expression("NumProj", IntType)
NumProjID = TriggerExpression("NumProjID", [IntType], IntType)
NumTarget = TriggerExpressionWithOptional("NumTarget", [IntType], IntType)
P1Name = Expression("P1Name", UStringType)
P2BodyDist = PositionExpression("P2BodyDist", FloatType)
P2Dist = PositionExpression("P2Dist", FloatType)
P2Life = Expression("P2Life", IntType)
P2MoveType = Expression("P2MoveType", defined.MoveTypeT)
P2StateNo = Expression("P2StateNo", IntType)
P2StateType = Expression("P2StateType", defined.StateTypeT)
P2Name = Expression("P2Name", UStringType)
P3Name = Expression("P3Name", UStringType)
P4Name = Expression("P4Name", UStringType)
PalNo = Expression("PalNo", IntType)
ParentDist = PositionExpression("ParentDist", FloatType)
Pi = Expression("pi", IntType)
Pos = PositionExpression("Pos", FloatType)
Power = Expression("Power", IntType)
PowerMax = Expression("PowerMax", IntType)
PlayerIDExist = TriggerExpression("PlayerIDExist", [IntType], IntType)
PrevStateNo = Expression("PrevStateNo", IntType)
ProjCancelTime = TriggerExpression("ProjCancelTime", [IntType], IntType)
ProjContactTime = TriggerExpression("ProjContactTime", [IntType], IntType)
ProjGuardedTime = TriggerExpression("ProjGuardedTime", [IntType], IntType)
ProjHitTime = TriggerExpression("ProjHitTime", [IntType], IntType)
Random = Expression("Random", IntType)
RootDist = PositionExpression("RootDist", FloatType)
RoundNo = Expression("RoundNo", IntType)
RoundsExisted = Expression("RoundsExisted", IntType)
RoundState = Expression("RoundState", IntType)
ScreenPos = PositionExpression("ScreenPos", FloatType)
SelfAnimExist = TriggerExpression("SelfAnimExist", [IntType], BoolType)
Sin = TriggerExpression("sin", [FloatType], FloatType)
StateNo = Expression("StateNo", IntType)
StateType = Expression("StateType", defined.StateTypeT)
StageVar = TriggerExpression("StageVar", [UStringType], UStringType)
## sysfvar
## sysvar
Tan = TriggerExpression("tan", [FloatType], FloatType)
TeamMode = Expression("TeamMode", defined.TeamModeTypeT)
TeamSide = Expression("TeamSide", IntType)
TicksPerSecond = Expression("TicksPerSecond", IntType)
Time = Expression("Time", IntType)
## TimeMod
## UniqHitCount
## var
Vel = PositionExpression("Vel", FloatType)
Win = Expression("Win", BoolType)
WinKO = Expression("WinKO", BoolType)
WinTime = Expression("WinTime", BoolType)
WinPerfect = Expression("WinPerfect", BoolType)

AnimElem = Expression("AnimElem", IntType)
ScreenWidth = Expression("ScreenWidth", IntType)
ScreenHeight = Expression("ScreenHeight", IntType)

EmptyExpression = Expression("", AnyType)
EmptyTuple = (EmptyExpression, )

__all__ = [
    "Abs", "Acos", "AiLevel", "Alive", "Anim", "AnimElemNo", "AnimElemTime", "AnimExist", "AnimTime", "Asin", "Atan",
    "AuthorName", "BackEdgeBodyDist", "BackEdgeDist", "CanRecover", "Ceil", "Command", "Const", "Const240p", "Const480p",
    "Const720p", "Cos", "Ctrl", "DrawGame", "E", "Exp", "Facing", "Floor", "FrontEdgeBodyDist", "FrontEdgeDist", "GameHeight",
    "GameTime", "GameWidth", "GetHitVar", "HitCount", "HitFall", "HitOver", "HitPauseTime", "HitShakeOver", "ID", "InGuardDist",
    "IsHelper", "IsHomeTeam", "Life", "LifeMax", "Ln", "Log", "Lose", "MatchNo", "MatchOver", "MoveContact", "MoveGuarded",
    "MoveHit", "MoveReversed", "Name", "NumEnemy", "NumExplod", "NumHelper", "NumPartner", "NumProj", "NumProjID", "NumTarget",
    "P1Name", "P2Life", "P2StateNo", "P2Name", "P3Name", "P4Name", "PalNo", "Pi", "Power", "PowerMax", "PlayerIDExist",
    "ProjCancelTime", "ProjContactTime", "ProjGuardedTime", "ProjHitTime", "Random", "RoundNo", "RoundsExisted", "RoundState",
    "SelfAnimExist", "Sin", "StateNo", "StageVar", "Tan", "TeamSide", "TicksPerSecond", "Time", "Win", "WinKO", "WinTime", "WinPerfect",
    "HitDefAttr", "StateType", "MoveType", "TeamMode", "P2MoveType", "P2StateType",
    "HitVel", "P2BodyDist", "P2Dist", "ParentDist", "Pos", "RootDist", "ScreenPos", "Vel",
    "ifelse", "Cond", "EmptyExpression", "EmptyTuple", "AnimElem", "ScreenWidth", "ScreenHeight"
]